theory Metatheory
imports Typing Semantics

begin

lemma weakening:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  and "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK"
  shows "\<Gamma> \<union> \<Gamma>' \<turnstile> e : \<tau>"
using prems by (induct arbitrary: \<Gamma>', auto simp add: wf_env_cons dom_def)

lemma substitution:
  fixes n :: nat
  assumes "size e = n"
  and "\<Gamma> \<turnstile> e' : \<tau>'"
  and "\<Gamma>,(x:\<tau>') \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  and "\<not> x free in e"
  shows "\<Gamma> \<turnstile> vsubst_tm e' i e : \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<Gamma> rule: nat_less_induct)
  case (1 n_y e_y \<tau>)
  have Hgok: "\<Gamma> \<Turnstile> OK" using prems by (auto simp add: ty_wf_env)
  show ?case proof (cases e_y)
    case (TmFreeVar y) thus ?thesis using Hgok prems by auto
  next
    case (TmVar j)
    show ?thesis proof (cases "j=i")
      case True
      hence Ht: "\<Gamma>,(x:\<tau>') \<turnstile> TmFreeVar x : \<tau>" using prems by auto
      hence "\<tau> = \<tau>'" using wf_env_def[of "\<Gamma>,(x:\<tau>')"] by fast
      thus ?thesis using True TmVar by auto
    next
      case False
      hence "\<Gamma>,(x:\<tau>') \<turnstile> TmVar j : \<tau>" using prems by auto
      thus ?thesis by (cases, auto)
    qed
  next
    case (TmAbs \<tau>_y e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) (Suc i) e_b"
    let ?e_b' = "vsubst_tm e' (Suc i) e_b"
    have Ht: "\<Gamma>,(x:\<tau>') \<turnstile> \<lambda>[\<tau>_y]. ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b L where Heq: "\<tau> = \<tau>_y \<rightarrow> \<tau>_b"
      and Hfin: "finite L"
      and Htyp: "\<forall>z. \<not> z free in ?e_bx \<and> z \<notin> L \<and> \<not> \<Gamma>,(x:\<tau>') defines z \<longrightarrow>
                       \<Gamma>,(x:\<tau>'),(z:\<tau>_y) \<turnstile> freshen_tm z ?e_bx : \<tau>_b" by auto
    let ?L = "L \<union> {x} \<union> set (free_tm ?e_bx)"
    have Hfin': "finite ?L" using Hfin by auto
    have Htyp': "\<forall>z. \<not> z free in ?e_b' \<and> z \<notin> ?L \<and> \<not> \<Gamma> defines z \<longrightarrow>
                   \<Gamma>,(z:\<tau>_y) \<turnstile> freshen_tm z ?e_b' : \<tau>_b" proof auto
      fix z
      assume Hnzf: "\<not> z free in ?e_b'"
        and Hnzl: "z \<notin> L"
        and Hnzbx: "\<not> z free in ?e_bx"
        and Hnzx: "z \<noteq> x"
        and Hnzg: "\<not> \<Gamma> defines z"
      have A1: "size (freshen_tm z e_b) < n_y" using prems by auto
      have "\<Gamma>,(z:\<tau>_y) \<Turnstile> OK" using Hgok Hnzg by (auto simp add: wf_env_def dom_def)
      hence A2: "\<Gamma>,(z:\<tau>_y) \<turnstile> e' : \<tau>'" using prems weakening[of _ _ _ "{(z\<rhd>\<tau>_y)}"] by auto
      have "\<not> z free in ?e_bx" and "z \<notin> L" and "\<not> \<Gamma>,(x:\<tau>') defines z"
        using Hnzbx Hnzl Hnzg Hnzx by (auto simp add: dom_def)
      hence "\<Gamma>,(x:\<tau>'),(z:\<tau>_y) \<turnstile> freshen_tm z ?e_bx : \<tau>_b" using Htyp by auto
      hence A3: "\<Gamma>,(z:\<tau>_y),(x:\<tau>') \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e_b) : \<tau>_b"
        by (auto simp add: env_permute subst_subst_swap[of "0" "Suc i"])
      have A4: "\<not> x free in (freshen_tm z e_b)" using prems by (auto simp add: not_free_subst)
      have "\<Gamma>,(z:\<tau>_y) \<turnstile> vsubst_tm e' (Suc i) (freshen_tm z e_b) : \<tau>_b"
        using "1.hyps" A1 A2 A3 A4 by auto
      moreover
      have "\<forall>k. vsubst_tm (TmFreeVar z) k e' = e'" using prems by (auto simp add: ty_same_subst)
      ultimately
      show "\<Gamma>,(z:\<tau>_y) \<turnstile> vsubst_tm (TmFreeVar z) 0 ?e_b' : \<tau>_b" by (auto simp add: subst_subst_swap)
    qed
    have "\<Gamma> \<turnstile> \<lambda>[\<tau>_y]. ?e_b' : \<tau>_y \<rightarrow> \<tau>_b" using Hgok Hfin' Htyp' ..
    thus ?thesis using TmAbs Heq by auto
  next
    case (TmApp e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmApp by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmApp by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    have "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x \<cdot> ?e2x : \<tau>" using prems by auto
    then obtain \<tau>1 where Ht1: "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x : \<tau>1 \<rightarrow> \<tau>"
      and Ht2: "\<Gamma>,(x:\<tau>') \<turnstile> ?e2x : \<tau>1" by auto
    have "\<Gamma> \<turnstile> vsubst_tm e' i e1 : \<tau>1 \<rightarrow> \<tau>" using "1.hyps" "1.prems" Hsize1 Hfree1 Ht1 by auto
    moreover
    have "\<Gamma> \<turnstile> vsubst_tm e' i e2 : \<tau>1" using "1.hyps" "1.prems" Hsize2 Hfree2 Ht2 by auto
    ultimately
    show ?thesis using TmApp by auto
  qed
qed

theorem preservation:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
    and   "e \<hookrightarrow> e'"
  shows   "\<Gamma> \<turnstile> e' : \<tau>"
using prems proof (induct arbitrary: e')
  case (t_var \<Gamma>_x \<tau>_x x e_x)
  have "TmFreeVar x \<hookrightarrow> e_x" .
  thus ?case by (cases, auto)
next
  case (t_abs L \<Gamma>_x \<tau>1_x \<tau>2_x e_b e_x)
  have "\<lambda>[\<tau>1_x]. e_b \<hookrightarrow> e_x" .
  thus ?case by (cases, auto)
next
  case (t_app \<Gamma>_x \<tau>1_x \<tau>2_x e1_x e2_x e_x)
  have "e1_x \<cdot> e2_x \<hookrightarrow> e_x" .
  thus ?case using t_app proof cases
    case (e_beta \<tau>1_y e1_y v2)
    hence "\<Gamma>_x \<turnstile> \<lambda>[\<tau>1_y]. e1_y : \<tau>1_x \<rightarrow> \<tau>2_x" using t_app by auto
    then obtain L where Hgok: "\<Gamma>_x \<Turnstile> OK" and Hfin: "finite L"
      and Hs: "\<forall>x. \<not> x free in e1_y \<and> x \<notin> L \<and> \<not> \<Gamma>_x defines x \<longrightarrow>
                 \<Gamma>_x,(x:\<tau>1_y) \<turnstile> freshen_tm x e1_y : \<tau>2_x" by (rule typing_abs_elim, auto)
    obtain z where Hz: "\<not> z free in e1_y \<and> z \<notin> L \<and> \<not> \<Gamma>_x defines z"
      using Hfin fresh_exists[of "set (free_tm e1_y) \<union> L \<union> dom \<Gamma>_x"]
        Hgok wf_env_def[of \<Gamma>_x] dom_finite[of \<Gamma>_x] by auto
    have "\<Gamma>_x,(z:\<tau>1_y) \<turnstile> freshen_tm z e1_y : \<tau>2_x" using Hs Hz by auto
    hence "\<Gamma>_x \<turnstile> vsubst_tm e2_x 0 e1_y : \<tau>2_x"
      using t_app e_beta Hz by (auto simp add: substitution)
    thus ?thesis using t_app e_beta by auto
  qed auto
qed

theorem progress:
  assumes "{} \<turnstile> e : \<tau>"
  shows "not_stuck e"
proof -
  { fix \<Gamma>
    assume "\<Gamma> \<turnstile> e : \<tau>" and "\<Gamma> = {}"
    hence "not_stuck e" proof induct
      case (t_app \<Gamma> \<tau>1 \<tau>2 e1 e2)
      hence "value e1 \<or> makes_progress e1" by auto
      thus ?case proof
        assume V1: "value e1"
        have "value e2 \<or> makes_progress e2" using t_app by auto
        thus ?case using V1 by auto
      next
        assume "makes_progress e1"
        thus "not_stuck (e1 \<cdot> e2)" by auto
      qed
    qed auto
  }
  thus ?thesis using prems by auto
qed

theorem program_safety:
  assumes Ht: "{} \<turnstile> e : \<tau>"
  and Hsem: "e \<hookrightarrow>* e'"
  shows "not_stuck e'"
using Hsem Ht proof induct
  case e_refl thus ?case using progress by auto
next
  case e_trans thus ?case using progress preservation by blast
qed

end
