theory Metatheory
imports Typing Semantics

begin

lemma weakening_m:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  and "M \<union> M' \<Turnstile> OK"
  shows "\<Gamma>;M \<union> M' \<turnstile> e : \<tau>"
using prems by (induct, auto)

lemma substitution:
  fixes n :: nat
  assumes "size e = n"
  and "\<Gamma>,(x:\<tau>');M \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  and "\<not> x free in e"
  and "\<not> \<Gamma> defines x"
  and "\<Gamma> \<sim> {(y\<rhd>\<tau>')}"
  and "\<Gamma>,(y:\<tau>') \<Turnstile> OK"
  shows "\<Gamma>,(y:\<tau>');M \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<Gamma> M rule: nat_less_induct)
  case (1 n_y e_y \<tau>)
  have Hgokx: "\<Gamma>,(x:\<tau>') \<Turnstile> OK" using prems ty_wf_env by auto
  have Hgoky: "\<Gamma>,(y:\<tau>') \<Turnstile> OK" .
  have Hmok: "M \<Turnstile> OK" using prems ty_wf_env_loc by auto
  show ?case proof (cases e_y)
    case (TmUnit q)
    hence Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmUnit q : \<tau>" using "1.prems" by auto
    hence Heq: "\<tau> = Qual q TyUnit" and "qualG \<Gamma>,(x:\<tau>') \<preceq> qU" by auto
    hence "qualG (\<Gamma>,(y:\<tau>')) \<preceq> qU" by (auto simp add: qualGrelL_def)
    hence "\<Gamma>,(y:\<tau>');M \<turnstile> TmUnit q : Qual q TyUnit" using Hmok Hgoky by auto
    thus ?thesis using TmUnit Heq by auto
  next
    case (TmFreeVar z)
    hence Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmFreeVar z : \<tau>" using "1.prems" by auto
    have Hneq: "z \<noteq> x" using prems by auto
    have "qualG ((\<Gamma>,(x:\<tau>'))\<bar>z) \<preceq> qU" using Ht by auto
    hence "qualT \<tau>' \<preceq> qU" and "qualG (\<Gamma>\<bar>z) \<preceq> qU"
      using Hneq by (auto simp add: without_def qualGrelL_def)
    hence "qualG (\<Gamma>,(y:\<tau>')\<bar>z) \<preceq> qU" by (cases "z=y", auto simp add: without_def qualGrelL_def)
    hence "\<Gamma>,(y:\<tau>');M \<turnstile> TmFreeVar z : \<tau>" using Ht Hneq Hmok Hgoky by auto
    thus ?thesis using TmFreeVar by auto
  next
    case (TmVar j)
    show ?thesis proof (cases "j=i")
      case True
      have Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmFreeVar x : \<tau>" using prems by auto
      hence Hteq: "\<tau> = \<tau>'" using wf_env_def[of "\<Gamma>,(x:\<tau>')"] by fast
      have "qualG ((\<Gamma>,(x:\<tau>'))\<bar>x) \<preceq> qU" using Ht by auto
      hence "qualG \<Gamma> \<preceq> qU" using "1.prems" by (auto simp add: dom_def without_def qualGrelL_def)
      hence "qualG (\<Gamma>\<bar>y) \<preceq> qU" using qualG_relL_subset[of \<Gamma> _ "\<Gamma>\<bar>y"] by (auto simp add: without_def)
      hence "qualG (\<Gamma>,(y:\<tau>')\<bar>y) \<preceq> qU" by (auto simp add: without_def qualGrelL_def)
      thus ?thesis using Hgoky Hmok Hteq TmVar True by auto
    next
      case False
      hence "\<Gamma>,(x:\<tau>');M \<turnstile> TmVar j : \<tau>" using prems by auto
      thus ?thesis by (cases, auto)
    qed
  next
    case (TmAbs q \<tau>_y e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) (Suc i) e_b"
    let ?e_by = "vsubst_tm (TmFreeVar y) (Suc i) e_b"
    have Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> \<lambda>^q[\<tau>_y]. ?e_bx : \<tau>" using prems by auto
    then obtain L \<tau>_b where Heq: "\<tau> = Qual q (\<tau>_y \<rightarrow> \<tau>_b)"
      and Hfin: "finite L"
      and Htyp: "\<forall>z. \<not> z free in ?e_bx \<and> z \<notin> L \<and> \<not> \<Gamma>,(x:\<tau>') defines z \<longrightarrow>
                      \<Gamma>,(x:\<tau>'),(z:\<tau>_y);M \<turnstile> freshen_tm z ?e_bx : \<tau>_b"
      and Hqual: "qualG (\<Gamma>,(x:\<tau>')) \<preceq> q" by auto
    let ?L = "L \<union> {x} \<union> set (free_tm ?e_bx)"
    have Hfin': "finite ?L" using Hfin by auto
    have Hqual': "qualG \<Gamma>,(y:\<tau>') \<preceq> q" using Hqual by (auto simp add: qualGrelL_def)
    have Htyp': "\<forall>z. \<not> z free in ?e_by \<and> z \<notin> ?L \<and> \<not> \<Gamma>,(y:\<tau>') defines z \<longrightarrow>
                   \<Gamma>,(y:\<tau>'),(z:\<tau>_y);M \<turnstile> freshen_tm z ?e_by : \<tau>_b" proof auto
      fix z
      assume Hnzby: "\<not> z free in ?e_by"
        and Hnzl: "z \<notin> L"
        and Hnzx: "z \<noteq> x"
        and Hnzbx: "\<not> z free in ?e_bx"
        and Hnzy: "z \<noteq> y"
        and Hnzg: "\<not> \<Gamma> defines z"
      have Hnx: "\<not> \<Gamma>,(x:\<tau>') defines z" using Hnzx Hnzg by (auto simp add: dom_def)
      have A1: "size (freshen_tm z e_b) < n_y" using prems by auto
      have "\<Gamma>,(x:\<tau>'),(z:\<tau>_y);M \<turnstile> freshen_tm z ?e_bx : \<tau>_b" using Htyp Hnzbx Hnzl Hnx by auto
      hence A2: "\<Gamma>,(z:\<tau>_y),(x:\<tau>');M \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e_b) : \<tau>_b"
        by (auto simp add: env_permute subst_subst_swap[of "Suc i" "0"])
      have A3: "\<not> x free in freshen_tm z e_b" using prems by (auto simp add: not_free_subst)
      have A4: "\<not> \<Gamma>,(z:\<tau>_y) defines x" using prems by (auto simp add: dom_def)
      have A5: "\<Gamma>,(z:\<tau>_y) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" Hnzy by (auto simp add: env_compat_def)
      have A6: "\<Gamma>,(z:\<tau>_y),(y:\<tau>') \<Turnstile> OK" using "1.prems" Hnzg Hnzy wf_env_cons[of "\<Gamma>,(y:\<tau>')" z \<tau>_y]
        by (auto simp add: env_permute)
      have "\<Gamma>,(z:\<tau>_y),(y:\<tau>');M \<turnstile> vsubst_tm (TmFreeVar y) (Suc i)
                                   (vsubst_tm (TmFreeVar z) 0 e_b) : \<tau>_b"
        using "1.hyps" A1 A2 A3 A4 A5 A6 by auto
      thus "\<Gamma>,(y:\<tau>'),(z:\<tau>_y);M \<turnstile> vsubst_tm (TmFreeVar z) 0 ?e_by : \<tau>_b"
        by (auto simp add: env_permute subst_subst_swap[of "Suc i" "0"])
    qed
    have "\<Gamma>,(y:\<tau>');M \<turnstile> \<lambda>^q[\<tau>_y]. ?e_by : Qual q (\<tau>_y \<rightarrow> \<tau>_b)" using Hgoky Hfin' Hqual' Hmok Htyp' ..
    thus ?thesis using TmAbs Heq by auto
  next
    case (TmApp e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmApp by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmApp by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    let ?e1y = "vsubst_tm (TmFreeVar y) i e1"
    let ?e2y = "vsubst_tm (TmFreeVar y) i e2"
    have "\<Gamma>,(x:\<tau>');M \<turnstile> ?e1x \<cdot> ?e2x : \<tau>" using prems by auto
    then obtain \<Gamma>1 \<Gamma>2 q \<tau>1 where Ht1: "\<Gamma>1;M \<turnstile> ?e1x : Qual q (\<tau>1 \<rightarrow> \<tau>)"
      and Ht2: "\<Gamma>2;M \<turnstile> ?e2x : \<tau>1"
      and Hcomp: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
      and Heq: "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" by auto
    have Hgok1: "\<Gamma>1 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>1] by auto
    have Hgok2: "\<Gamma>2 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>2] by auto
    have Hsub1x: "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp1x: "(\<Gamma>1\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>1\<bar>x"] by auto
    have Hsub2x: "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp2x: "(\<Gamma>2\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>2\<bar>x"] by auto
    have "(\<Gamma>1\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub1x by auto
    hence Hgok1xy: "(\<Gamma>1\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>1\<bar>x),(y:\<tau>')"] by auto
    have "(\<Gamma>2\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub2x by auto
    hence Hgok2xy: "(\<Gamma>2\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>2\<bar>x),(y:\<tau>')"] by auto
    have Lnx1: "\<not> \<Gamma>1 defines x \<longrightarrow> \<Gamma>1;M \<turnstile> ?e1y : Qual q (\<tau>1 \<rightarrow> \<tau>)"
      using Ht1 by (auto simp add: not_in_env_ty_same_subst)
    have Lnx2: "\<not> \<Gamma>2 defines x \<longrightarrow> \<Gamma>2;M \<turnstile> ?e2y : \<tau>1"
      using Ht2 by (auto simp add: not_in_env_ty_same_subst)
    have Ldx1: "\<Gamma>1 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual q (\<tau>1 \<rightarrow> \<tau>)" proof
      assume "\<Gamma>1 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        using "1.prems" TmApp Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>1 x \<tau>'] by auto
      hence "(\<Gamma>1\<bar>x),(x:\<tau>') = \<Gamma>1" using Hgok1 wf_env_defines_without by auto
      also note Ht1
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual q (\<tau>1 \<rightarrow> \<tau>)"
        using H "1.hyps" Hsize1 Hfree1 Hcomp1x Hgok1xy by (auto simp add: without_def dom_def)
    qed
    have Ldx2: "\<Gamma>2 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>1" proof
      assume "\<Gamma>2 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
        using "1.prems" TmApp Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>2 x \<tau>'] by auto
      hence "(\<Gamma>2\<bar>x),(x:\<tau>') = \<Gamma>2" using Hgok2 wf_env_defines_without by auto
      also note Ht2
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>1"
        using H "1.hyps" Hsize2 Hfree2 Hcomp2x Hgok2xy by (auto simp add: without_def dom_def)
    qed
    have "\<Gamma>,(y:\<tau>');M \<turnstile> ?e1y \<cdot> ?e2y : \<tau>" proof (cases "\<Gamma>1 defines x")
      case True
      hence Hxt1: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        and Ht1': "(\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual q (\<tau>1 \<rightarrow> \<tau>)" using Ldx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>1" using Ldx2 by auto
        have H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')"
          using "1.prems" Heq by (auto simp add: without_def dom_def)
        have "qualT \<tau>' \<preceq> qU" using Hxt1 Hxt2 Hcomp by (auto simp add: env_compat_def)
        hence "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')"
          using Hcomp by (auto simp add: env_compat_def without_def)
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y \<cdot> ?e2y : \<tau>" using Ht1' Ht2' ..
        thus ?thesis using H by auto
      next
        case False
        hence Ht2': "\<Gamma>2;M \<turnstile> ?e2y : \<tau>1" using Lnx2 by auto
        have "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" using "1.prems" Heq False
        proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>2" and "\<forall>b. (x, b) \<notin> \<Gamma>2"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 = \<Gamma>,(y:\<tau>')" by auto
        have "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> \<Gamma>2" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>2"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2;M \<turnstile> ?e1y \<cdot> ?e2y : \<tau>" using Ht1' Ht2' ..
        thus ?thesis using H by auto
      qed
    next
      case False
      hence Ht1': "\<Gamma>1;M \<turnstile> ?e1y : Qual q (\<tau>1 \<rightarrow> \<tau>)" using Lnx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>1" using Ldx2 by auto
        have "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" using "1.prems" Heq False proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>1" and "\<forall>b. (x, b) \<notin> \<Gamma>1"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')" by auto
        have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>1"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y \<cdot> ?e2y : \<tau>" using Ht1' Ht2' ..
        thus ?thesis using H by auto
      next
        case False
        have "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" using prems by (auto simp add: dom_def)
        hence "\<not> \<Gamma>,(x:\<tau>') defines x" using Heq by auto
        thus ?thesis by (auto simp add: dom_def)
      qed
    qed
    thus ?thesis using TmApp by auto
  next
    case (TmNew e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    let ?e_by = "vsubst_tm (TmFreeVar y) i e_b"
    have Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmNew ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b where Heq: "\<tau> = Qual qL (TyRef \<tau>_b)" and Htb: "\<Gamma>,(x:\<tau>');M \<turnstile> ?e_bx : \<tau>_b" by auto
    have "\<Gamma>,(y:\<tau>');M \<turnstile> ?e_by : \<tau>_b" using "1.hyps" TmNew Htb "1.prems" by auto
    thus ?thesis using TmNew Heq by auto
  next
    case (TmFree e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    let ?e_by = "vsubst_tm (TmFreeVar y) i e_b"
    have Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmFree ?e_bx : \<tau>" using prems by auto
    hence "\<Gamma>,(x:\<tau>');M \<turnstile> ?e_bx : Qual qL (TyRef \<tau>)" by auto
    hence "\<Gamma>,(y:\<tau>');M \<turnstile> ?e_by : Qual qL (TyRef \<tau>)" using "1.hyps" TmFree "1.prems" by auto
    thus ?thesis using TmFree by auto
  next
    case (TmDeref e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    let ?e_by = "vsubst_tm (TmFreeVar y) i e_b"
    have Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmDeref ?e_bx : \<tau>" using prems by auto
    hence "\<Gamma>,(x:\<tau>');M \<turnstile> ?e_bx : Qual qU (TyRef \<tau>)" and Hq: "qualT \<tau> \<preceq> qU" by auto
    hence "\<Gamma>,(y:\<tau>');M \<turnstile> ?e_by : Qual qU (TyRef \<tau>)" using "1.hyps" TmDeref "1.prems" by auto
    thus ?thesis using TmDeref Hq by auto
  next
    case (TmAssign e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmAssign by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmAssign by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    let ?e1y = "vsubst_tm (TmFreeVar y) i e1"
    let ?e2y = "vsubst_tm (TmFreeVar y) i e2"
    have "\<Gamma>,(x:\<tau>');M \<turnstile> ?e1x := ?e2x : \<tau>" using prems by auto
    then obtain \<Gamma>1 \<Gamma>2 \<tau>_b where Ht1: "\<Gamma>1;M \<turnstile> ?e1x : Qual qU (TyRef \<tau>_b)"
      and Ht2: "\<Gamma>2;M \<turnstile> ?e2x : \<tau>_b" and Hqual: "qualT \<tau>_b \<preceq> qU"
      and Hcomp: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
      and Heq: "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and Heqt: "\<tau> = Qual qU TyUnit" by auto
    have Hgok1: "\<Gamma>1 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>1] by auto
    have Hgok2: "\<Gamma>2 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>2] by auto
    have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>')" using Heq by auto
    have Hsub1x: "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp1x: "(\<Gamma>1\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>1\<bar>x"] by auto
    have Hsub2x: "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp2x: "(\<Gamma>2\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>2\<bar>x"] by auto
    have "(\<Gamma>1\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub1x by auto
    hence Hgok1xy: "(\<Gamma>1\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>1\<bar>x),(y:\<tau>')"] by auto
    have "(\<Gamma>2\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub2x by auto
    hence Hgok2xy: "(\<Gamma>2\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>2\<bar>x),(y:\<tau>')"] by auto
    have Lnx1: "\<not> \<Gamma>1 defines x \<longrightarrow> \<Gamma>1;M \<turnstile> ?e1y : Qual qU (TyRef \<tau>_b)"
      using Ht1 by (auto simp add: not_in_env_ty_same_subst)
    have Lnx2: "\<not> \<Gamma>2 defines x \<longrightarrow> \<Gamma>2;M \<turnstile> ?e2y : \<tau>_b"
      using Ht2 by (auto simp add: not_in_env_ty_same_subst)
    have Ldx1: "\<Gamma>1 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>_b)" proof
      assume "\<Gamma>1 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        using "1.prems" TmAssign Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>1 x \<tau>'] by auto
      hence "(\<Gamma>1\<bar>x),(x:\<tau>') = \<Gamma>1" using Hgok1 wf_env_defines_without by auto
      also note Ht1
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>_b)"
        using H "1.hyps" Hsize1 Hfree1 Hcomp1x Hgok1xy by (auto simp add: without_def dom_def)
    qed
    have Ldx2: "\<Gamma>2 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>_b" proof
      assume "\<Gamma>2 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
        using "1.prems" TmAssign Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>2 x \<tau>'] by auto
      hence "(\<Gamma>2\<bar>x),(x:\<tau>') = \<Gamma>2" using Hgok2 wf_env_defines_without by auto
      also note Ht2
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>_b"
        using H "1.hyps" Hsize2 Hfree2 Hcomp2x Hgok2xy by (auto simp add: without_def dom_def)
    qed
    have "\<Gamma>,(y:\<tau>');M \<turnstile> ?e1y := ?e2y : \<tau>" proof (cases "\<Gamma>1 defines x")
      case True
      hence Hxt1: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        and Ht1': "(\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>_b)" using Ldx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>_b" using Ldx2 by auto
        have H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')"
          using "1.prems" Heq by (auto simp add: without_def dom_def)
        have "qualT \<tau>' \<preceq> qU" using Hxt1 Hxt2 Hcomp by (auto simp add: env_compat_def)
        hence "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')"
          using Hcomp by (auto simp add: env_compat_def without_def)
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y := ?e2y : Qual qU TyUnit"
          using Hqual Ht1' Ht2' ..
        thus ?thesis using H Heqt by auto
      next
        case False
        hence Ht2': "\<Gamma>2;M \<turnstile> ?e2y : \<tau>_b" using Lnx2 by auto
        have "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" using "1.prems" Heq False
        proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>2" and "\<forall>b. (x, b) \<notin> \<Gamma>2"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 = \<Gamma>,(y:\<tau>')" by auto
        have "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> \<Gamma>2" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>2"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2;M \<turnstile> ?e1y := ?e2y : Qual qU TyUnit" using Hqual Ht1' Ht2' ..
        thus ?thesis using H Heqt by auto
      qed
    next
      case False
      hence Ht1': "\<Gamma>1;M \<turnstile> ?e1y : Qual qU (TyRef \<tau>_b)" using Lnx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>_b" using Ldx2 by auto
        have "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" using "1.prems" Heq False proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>1" and "\<forall>b. (x, b) \<notin> \<Gamma>1"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')" by auto
        have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>1"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y := ?e2y : Qual qU TyUnit" using Hqual Ht1' Ht2' ..
        thus ?thesis using H Heqt by auto
      next
        case False
        have "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" using prems by (auto simp add: dom_def)
        hence "\<not> \<Gamma>,(x:\<tau>') defines x" using Heq by auto
        thus ?thesis by (auto simp add: dom_def)
      qed
    qed
    thus ?thesis using TmAssign by auto
  next
    case (TmSwap e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmSwap by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmSwap by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    let ?e1y = "vsubst_tm (TmFreeVar y) i e1"
    let ?e2y = "vsubst_tm (TmFreeVar y) i e2"
    have "\<Gamma>,(x:\<tau>');M \<turnstile> ?e1x :=: ?e2x : \<tau>" using prems by auto
    then obtain \<Gamma>1 \<Gamma>2 where Ht1: "\<Gamma>1;M \<turnstile> ?e1x : Qual qU (TyRef \<tau>)"
      and Ht2: "\<Gamma>2;M \<turnstile> ?e2x : \<tau>" and Hqual: "qL \<preceq> qualT \<tau>"
      and Hcomp: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
      and Heq: "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" by auto
    have Hgok1: "\<Gamma>1 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>1] by auto
    have Hgok2: "\<Gamma>2 \<Turnstile> OK" using "1.prems" Heq Hgokx wf_env_subset[of "\<Gamma>,(x:\<tau>')" \<Gamma>2] by auto
    have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>')" using Heq by auto
    have Hsub1x: "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp1x: "(\<Gamma>1\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>1\<bar>x"] by auto
    have Hsub2x: "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using Heq by (auto simp add: without_def)
    hence Hcomp2x: "(\<Gamma>2\<bar>x) \<sim> {(y\<rhd>\<tau>')}" using "1.prems" env_compat_subset[of \<Gamma> _ "\<Gamma>2\<bar>x"] by auto
    have "(\<Gamma>1\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub1x by auto
    hence Hgok1xy: "(\<Gamma>1\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>1\<bar>x),(y:\<tau>')"] by auto
    have "(\<Gamma>2\<bar>x),(y:\<tau>') \<subseteq> \<Gamma>,(y:\<tau>')" using Hsub2x by auto
    hence Hgok2xy: "(\<Gamma>2\<bar>x),(y:\<tau>') \<Turnstile> OK"
      using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>')" "(\<Gamma>2\<bar>x),(y:\<tau>')"] by auto
    have Lnx1: "\<not> \<Gamma>1 defines x \<longrightarrow> \<Gamma>1;M \<turnstile> ?e1y : Qual qU (TyRef \<tau>)"
      using Ht1 by (auto simp add: not_in_env_ty_same_subst)
    have Lnx2: "\<not> \<Gamma>2 defines x \<longrightarrow> \<Gamma>2;M \<turnstile> ?e2y : \<tau>"
      using Ht2 by (auto simp add: not_in_env_ty_same_subst)
    have Ldx1: "\<Gamma>1 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>)" proof
      assume "\<Gamma>1 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        using "1.prems" TmSwap Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>1 x \<tau>'] by auto
      hence "(\<Gamma>1\<bar>x),(x:\<tau>') = \<Gamma>1" using Hgok1 wf_env_defines_without by auto
      also note Ht1
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>1 \<and> (\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>)"
        using H "1.hyps" Hsize1 Hfree1 Hcomp1x Hgok1xy by (auto simp add: without_def dom_def)
    qed
    have Ldx2: "\<Gamma>2 defines x \<longrightarrow> (x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>" proof
      assume "\<Gamma>2 defines x"
      hence H: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
        using "1.prems" TmSwap Heq Hgokx wf_env_subset_defines[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>2 x \<tau>'] by auto
      hence "(\<Gamma>2\<bar>x),(x:\<tau>') = \<Gamma>2" using Hgok2 wf_env_defines_without by auto
      also note Ht2
      finally
      show "(x\<rhd>\<tau>') \<in> \<Gamma>2 \<and> (\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>"
        using H "1.hyps" Hsize2 Hfree2 Hcomp2x Hgok2xy by (auto simp add: without_def dom_def)
    qed
    have "\<Gamma>,(y:\<tau>');M \<turnstile> ?e1y :=: ?e2y : \<tau>" proof (cases "\<Gamma>1 defines x")
      case True
      hence Hxt1: "(x\<rhd>\<tau>') \<in> \<Gamma>1"
        and Ht1': "(\<Gamma>1\<bar>x),(y:\<tau>');M \<turnstile> ?e1y : Qual qU (TyRef \<tau>)" using Ldx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>" using Ldx2 by auto
        have H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')"
          using "1.prems" Heq by (auto simp add: without_def dom_def)
        have "qualT \<tau>' \<preceq> qU" using Hxt1 Hxt2 Hcomp by (auto simp add: env_compat_def)
        hence "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')"
          using Hcomp by (auto simp add: env_compat_def without_def)
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y :=: ?e2y : \<tau>"
          using Hqual Ht1' Ht2' ..
        thus ?thesis using H by auto
      next
        case False
        hence Ht2': "\<Gamma>2;M \<turnstile> ?e2y : \<tau>" using Lnx2 by auto
        have "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" using "1.prems" Heq False
        proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>2" and "\<forall>b. (x, b) \<notin> \<Gamma>2"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 = \<Gamma>,(y:\<tau>')" by auto
        have "(\<Gamma>1\<bar>x),(y:\<tau>') \<sim> \<Gamma>2" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>2"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2 \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "((\<Gamma>1\<bar>x),(y:\<tau>')) \<union> \<Gamma>2;M \<turnstile> ?e1y :=: ?e2y : \<tau>" using Hqual Ht1' Ht2' ..
        thus ?thesis using H by auto
      qed
    next
      case False
      hence Ht1': "\<Gamma>1;M \<turnstile> ?e1y : Qual qU (TyRef \<tau>)" using Lnx1 by auto
      show ?thesis proof (cases "\<Gamma>2 defines x")
        case True
        hence Hxt2: "(x\<rhd>\<tau>') \<in> \<Gamma>2"
          and Ht2': "(\<Gamma>2\<bar>x),(y:\<tau>');M \<turnstile> ?e2y : \<tau>" using Ldx2 by auto
        have "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" using "1.prems" Heq False proof (auto simp add: dom_def without_def)
          fix a b
          assume "\<Gamma>,(x:\<tau>') = \<Gamma>1 \<union> \<Gamma>2" and "(a, b) \<in> \<Gamma>1" and "\<forall>b. (x, b) \<notin> \<Gamma>1"
          thus "(a, b) \<in> \<Gamma>" by (cases "a = x", auto)
        qed
        hence H: "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) = \<Gamma>,(y:\<tau>')" by auto
        have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>')" using Hcomp
        proof (auto simp add: env_compat_def without_def dom_def)
          assume "(y\<rhd>\<tau>') \<in> \<Gamma>1"
          hence "(y\<rhd>\<tau>') \<in> \<Gamma>" using False Heq by (cases "y = x", auto simp add: dom_def)
          thus "qualT \<tau>' \<preceq> qU" using "1.prems" by (auto simp add: env_compat_def)
        qed
        moreover
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>')) \<Turnstile> OK" using H "1.prems" by auto
        ultimately
        have "\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>'));M \<turnstile> ?e1y :=: ?e2y : \<tau>" using Hqual Ht1' Ht2' ..
        thus ?thesis using H by auto
      next
        case False
        have "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" using prems by (auto simp add: dom_def)
        hence "\<not> \<Gamma>,(x:\<tau>') defines x" using Heq by auto
        thus ?thesis by (auto simp add: dom_def)
      qed
    qed
    thus ?thesis using TmSwap by auto
  next
    case (TmLoc q l)
    hence Ht: "\<Gamma>,(x:\<tau>');M \<turnstile> TmLoc q l : \<tau>" using prems by auto
    then obtain \<tau>_b where Heq: "\<tau> = Qual q (TyRef \<tau>_b)"
      and Hin: "(l\<rhd>\<tau>_b) \<in> M"
      and Hqual: "qualG \<Gamma>,(x:\<tau>') \<preceq> qU" by auto
    have "qualG \<Gamma>,(y:\<tau>') \<preceq> qU" using Hqual by (auto simp add: qualGrelL_def)
    hence "\<Gamma>,(y:\<tau>');M \<turnstile> TmLoc q l : Qual q (TyRef \<tau>_b)" using Hgoky Hmok Heq Hin by auto
    thus ?thesis using TmLoc Heq by auto
  qed
qed

consts storeTyping :: "(env * store * env) set"
syntax
  "_storeTyping" :: "env \<Rightarrow> store \<Rightarrow> env \<Rightarrow> bool"      ("_ \<Turnstile> _:_" [50,50,50] 50)
translations
  "M \<Turnstile> S:\<Gamma>" \<rightleftharpoons> "(M, S, \<Gamma>) \<in> storeTyping"

inductive storeTyping
  intros
  st_nil[intro!]   : "M \<Turnstile> {} : {}"
  st_cons[intro!]  : "\<lbrakk> \<Gamma>v;M \<turnstile> v : \<tau>; M \<Turnstile> S : \<Gamma>s \<union> \<Gamma>v; \<Gamma>s \<sim> \<Gamma>v; S \<Turnstile> Store;
                        \<Gamma>s \<union> \<Gamma>v \<Turnstile> OK; value v; \<not> S defines x; un \<Gamma>v \<subseteq> un \<Gamma>s \<rbrakk> \<Longrightarrow>
                        M \<Turnstile> S,(x\<mapsto>v) : \<Gamma>s,(x:\<tau>)"

inductive_cases st_nil_elim [elim!]:  "M \<Turnstile> {} : {}"
inductive_cases st_cons_elim [elim!]: "M \<Turnstile> S,(x\<mapsto>v) : \<Gamma>,(x:\<tau>)"

consts memTyping :: "(env * store * env) set"
syntax
  "_memTyping" :: "env \<Rightarrow> store \<Rightarrow> env \<Rightarrow> bool"      ("_ \<Turnstile>' _:_" [50,50,50] 50)
translations
  "\<Gamma> \<Turnstile>' \<mu>:M" \<rightleftharpoons> "(\<Gamma>, \<mu>, M) \<in> memTyping"

inductive memTyping
  intros
  mt_nil[intro!]   : "\<Gamma> \<Turnstile>' {} : {}"
  mt_cons[intro!]  : "\<lbrakk> \<Gamma> \<Turnstile>' \<mu> : M; \<Gamma> \<sim> {x\<rhd>\<tau>}; \<Gamma>,(x:\<tau>) \<Turnstile> OK; \<not> \<mu> defines l; M \<Turnstile> OK; \<mu> \<Turnstile> Mem \<rbrakk> \<Longrightarrow>
                        \<Gamma>,(x:\<tau>) \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x) : M,(l:\<tau>)"

inductive_cases mt_nil_elim [elim!]:  "\<Gamma> \<Turnstile>' {} : {}"
inductive_cases mt_cons_elim [elim!]: "\<Gamma> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x) : M,(x:\<tau>)"

consts
  INV2 :: "store \<Rightarrow> bool"
  INV3 :: "store \<Rightarrow> bool"
defs
  INV2[simp]: "INV2(S) \<equiv> \<forall>x l. (x\<mapsto>TmLoc qL l) \<in> S  \<longrightarrow> \<not> l free loc in st (S\<bar>x)"
  INV3[simp]: "INV3(S) \<equiv> \<forall>x v. (x\<mapsto>v) \<in> S \<longrightarrow> (\<exists>q l. v = TmLoc q l) \<or> (\<forall>l. \<not> l free loc in v)"

lemma inv2_subset:
  assumes "INV2 S"
  and "S' \<subseteq> S"
  shows "INV2 S'"
using prems proof auto
  fix x l
  assume "(x\<mapsto>TmLoc qL l) \<in> S'" and "l free loc in st (S'\<bar>x)"
  have "(S'\<bar>x) \<subseteq> (S\<bar>x)" using prems by (auto simp add: without_def)
  hence "l free loc in st (S\<bar>x)" using prems by (auto simp add: free_loc_store_def)
  thus False using prems by auto
qed

lemma inv3_subset:
  assumes "INV3 S"
  and "S' \<subseteq> S"
  shows "INV3 S'"
using prems proof auto
  fix x v
  assume "(x\<mapsto>v) \<in> S'" and "free_loc v \<noteq> []"
  have "(x\<mapsto>v) \<in> S" using prems by auto
  thus "\<exists>q l. v = TmLoc q l" using prems by auto
qed

lemma inv2_cons:
  assumes "INV2 S"
  and "\<not> S defines z"
  and "\<forall>l. \<not> l free loc in v"
  shows "INV2 (S,(z\<mapsto>v))"
using prems proof auto
  fix x l
  assume "(x\<mapsto>TmLoc qL l) \<in> S" and "l free loc in st (S,(z\<mapsto>v) \<bar> x)"
  hence "\<not> l free loc in st (S\<bar>x)" using prems by auto
  moreover
  have "\<not> l free loc in v" using prems by auto
  ultimately
  have "\<not> l free loc in st ((S\<bar>x),(z\<mapsto>v))" using free_loc_store_def by auto
  also
  have "z \<noteq> x" using prems by (auto simp add: dom_def)
  hence "(S\<bar>x),(z\<mapsto>v) = (S,(z\<mapsto>v))\<bar>x" by (auto simp add: without_def)
  finally show False using prems by auto
qed

lemma inv2_cons_loc:
  assumes "INV2 S"
  and "\<not> S defines z"
  and "\<not> l free loc in st S"
  shows "INV2 (S,(z\<mapsto>TmLoc qL l))"
using prems proof auto
  assume "l free loc in st (S,(z\<mapsto>TmLoc qL l)\<bar>z)"
  have "S,(z\<mapsto>TmLoc qL l)\<bar>z = S" using prems by (auto simp add: without_def dom_def)
  thus False using prems by auto
next
  fix x lx
  assume "(x\<mapsto>TmLoc qL lx) \<in> S" and "lx free loc in st (S,(z\<mapsto>TmLoc qL l)\<bar>x)"
  hence "S,(z\<mapsto>TmLoc qL l)\<bar>x = (S\<bar>x),(z\<mapsto>TmLoc qL l)"
    using prems by (auto simp add: without_def dom_def)
  hence "lx free loc in st ((S\<bar>x),(z\<mapsto>TmLoc qL l))" using prems by auto
  moreover
  hence "\<not> lx free loc in st (S\<bar>x)" using prems by auto
  ultimately
  have Heq: "l = lx" by (auto simp add: free_loc_store_def)
  have "lx free loc in st S" proof (auto simp add: free_loc_store_def)
    show "\<exists>x. (\<exists>z v. x = set (free_loc v) \<and> (z\<mapsto>v) \<in> S) \<and> lx \<in> x" proof (rule exI | rule conjI)+
      show "(x\<mapsto>TmLoc qL lx) \<in> S" using prems by auto
    qed auto
  qed
  thus False using Heq prems by auto
qed

lemma store_dom_in:
  assumes "M \<Turnstile> S : \<Gamma>"
  and "\<Gamma> defines z"
  shows "S defines z"
using prems proof induct
  case (st_cons M S \<Gamma>s \<Gamma>v \<tau> v x)
  thus ?case by (cases "x=z", auto simp add: dom_def)
qed (auto simp add: dom_def)

lemma store_dom_not_in:
  assumes "M \<Turnstile> S : \<Gamma>"
  and "\<not> S defines z"
  shows "\<not> \<Gamma> defines z"
using prems by (auto simp only: store_dom_in)

lemma storeTyping_weakening:
  assumes "M \<Turnstile> S : \<Gamma>s"
  and "M,(l:\<tau>) \<Turnstile> OK"
  shows "M,(l:\<tau>) \<Turnstile> S : \<Gamma>s"
using prems proof induct
  case (st_cons M S \<Gamma>s \<Gamma>v \<tau>z vz z)
  have "\<Gamma>v;M,(l:\<tau>) \<turnstile> vz : \<tau>z" using prems weakening_m[of _ _ _ _ "{l\<rhd>\<tau>}"] by auto
  thus ?case using st_cons by fast
qed auto

lemma mem_dom_in:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "M defines l"
  shows "\<mu> defines l"
using prems proof induct
  case (mt_cons M \<Gamma> \<mu> \<tau> z x)
  thus ?case by (cases "z=l", auto simp add: dom_def)
qed (auto simp add: dom_def)

lemma mem_dom_not_in:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "\<not> \<mu> defines l"
  shows "\<not> M defines l"
using prems by (auto simp add: mem_dom_in)

lemma memTyping_weakening:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "\<Gamma>,(x:\<tau>) \<Turnstile> OK"
  and "\<Gamma> \<sim> {x\<rhd>\<tau>}"
  shows "\<Gamma>,(x:\<tau>) \<Turnstile>' \<mu> : M"
using prems proof induct
  case (mt_cons M \<Gamma> \<mu> \<tau>z lz z)
  thus ?case proof (cases "z = x")
    case True
    have "\<tau> = \<tau>z" using prems wf_env_def[of "\<Gamma>,(z:\<tau>z),(x:\<tau>)"] by auto
    thus ?thesis using prems by auto
  next
    case False
    have "\<Gamma>,(x:\<tau>),(z:\<tau>z) \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar z):M,(lz:\<tau>z)" using prems proof auto
      have "\<Gamma>,(x:\<tau>) \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>,(z:\<tau>z),(x:\<tau>)" "\<Gamma>,(x:\<tau>)"] by auto
      moreover
      have "\<Gamma> \<sim> {(x\<rhd>\<tau>)}" using prems by (auto simp add: env_compat_def)
      ultimately
      show "\<Gamma>,(x:\<tau>) \<Turnstile>' \<mu>:M" using prems by auto
    next
      show "\<Gamma>,(x:\<tau>) \<sim> {(z, Term \<tau>z)}" using prems by (auto simp add: env_compat_def)
    next
      show "\<Gamma>,(x:\<tau>),(z:\<tau>z) \<Turnstile> OK" using prems by (auto simp add: env_permute)
    qed      
    thus ?thesis by (auto simp add: env_permute)
  qed
qed auto

lemma memTyping_elim_qU:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "(l\<rhd>\<tau>) \<in> M"
  and "qualT \<tau> = qU"
  shows "\<Gamma> \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (x\<rhd>\<tau>) \<in> \<Gamma>"
using prems proof induct
  case (mt_cons M \<Gamma> \<mu> \<tau>z lz xz)
  hence Hnotinm: "\<not> M defines lz" using mem_dom_not_in by auto
  hence Hmok: "M,(lz:\<tau>z) \<Turnstile> OK" using mt_cons wf_env_cons by auto
  have Hmt: "\<mu>,(lz\<mapsto>TmFreeVar xz) \<Turnstile> Mem" using mt_cons wf_mem_cons by auto
  show ?case proof (cases "l = lz")
    case True
    hence Hteq: "\<tau> = \<tau>z" using Hmok mt_cons wf_env_def[of "M,(lz:\<tau>z)"] by fast
    have Hxeq: "x = xz" using Hmt True mt_cons wf_mem_def[of "\<mu>,(lz\<mapsto>TmFreeVar xz)"] by fast
    have Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>,(xz:\<tau>z)" using Hteq Hxeq by auto
    have "\<Gamma> \<Turnstile>' \<mu>:M" by assumption
    also have "\<mu> = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l" using mt_cons True by (auto simp add: without_def dom_def)
    also have "M = M,(lz:\<tau>z)\<bar>l" using Hnotinm True by (auto simp add: without_def dom_def)
    finally have "\<Gamma>,(xz:\<tau>z) \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l : M,(lz:\<tau>z)\<bar>l"
      using memTyping_weakening mt_cons by auto
    thus ?thesis using Hin by auto
  next
    case False
    hence "(l\<mapsto>TmFreeVar x) \<in> \<mu>" and "(l\<rhd>\<tau>) \<in> M" and "qualT \<tau> = qU" using prems by auto
    hence Hmt': "\<Gamma> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" and Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>" using mt_cons.hyps by auto
    have "\<Gamma>,(xz:\<tau>z) \<Turnstile>' (\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) : (M\<bar>l),(lz:\<tau>z)" proof
      show "\<Gamma> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" using Hmt' by auto
    next
      show "\<not> \<mu>\<bar>l defines lz" using mt_cons False by (auto simp add: dom_def without_def)
    next
      show "M\<bar>l \<Turnstile> OK" using mt_cons wf_env_subset[of M "M\<bar>l"] by (auto simp add: without_def)
    next
      show "\<mu>\<bar>l \<Turnstile> Mem" using wf_mem_subset[of \<mu> "\<mu>\<bar>l"] mt_cons by (auto simp add: without_def)
    qed
    also have "(\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l"
      using False by (auto simp add: without_def)
    also have "(M\<bar>l),(lz:\<tau>z) = M,(lz:\<tau>z)\<bar>l" using False by (auto simp add: without_def)
    finally show ?thesis using Hin by auto
  qed
qed auto

lemma memTyping_elim_qL:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "(l\<rhd>\<tau>) \<in> M"
  and "qualT \<tau> = qL"
  shows "\<Gamma>\<bar>x \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (x\<rhd>\<tau>) \<in> \<Gamma>"
using prems proof induct
  case (mt_cons M \<Gamma> \<mu> \<tau>z lz xz)
  hence Hnotinm: "\<not> M defines lz" using mem_dom_not_in by auto
  hence Hmok: "M,(lz:\<tau>z) \<Turnstile> OK" using mt_cons wf_env_cons by auto
  have Hmt: "\<mu>,(lz\<mapsto>TmFreeVar xz) \<Turnstile> Mem" using mt_cons wf_mem_cons by auto
  show ?case proof (cases "l = lz")
    case True
    hence Hteq: "\<tau> = \<tau>z" using Hmok mt_cons wf_env_def[of "M,(lz:\<tau>z)"] by fast
    have Hxeq: "x = xz" using Hmt True mt_cons wf_mem_def[of "\<mu>,(lz\<mapsto>TmFreeVar xz)"] by fast
    have Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>,(xz:\<tau>z)" using Hteq Hxeq by auto
    have "\<Gamma> \<Turnstile>' \<mu>:M" by assumption
    also have "\<mu> = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l" using mt_cons True by (auto simp add: without_def dom_def)
    also have "M = M,(lz:\<tau>z)\<bar>l" using Hnotinm True by (auto simp add: without_def dom_def)
    also have "\<not> \<Gamma> defines xz" using Hteq compat_ok_sanity_right[of \<Gamma> "{xz\<rhd>\<tau>z}" xz \<tau>z] mt_cons
      by auto
    hence "\<Gamma> = \<Gamma>,(xz:\<tau>z)\<bar>x" using Hxeq by (auto simp add: without_def dom_def)
    finally have "\<Gamma>,(xz:\<tau>z)\<bar>x \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l : M,(lz:\<tau>z)\<bar>l" 
      using memTyping_weakening mt_cons by auto
    thus ?thesis using Hin by auto
  next
    case False
    hence "(l\<mapsto>TmFreeVar x) \<in> \<mu>" and "(l\<rhd>\<tau>) \<in> M" using prems by auto
    hence Hmt': "\<Gamma>\<bar>x \<Turnstile>' \<mu>\<bar>l : M\<bar>l" and Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>" using mt_cons by auto
    hence Hxneq: "xz \<noteq> x"
      using mt_cons compat_ok_sanity[of \<Gamma> "{xz\<rhd>\<tau>z}" x \<tau>] by (auto simp add: dom_def)
    have "(\<Gamma>\<bar>x),(xz:\<tau>z) \<Turnstile>' (\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) : (M\<bar>l),(lz:\<tau>z)" proof
      show "\<Gamma>\<bar>x \<Turnstile>' \<mu>\<bar>l : M\<bar>l" using Hmt' by auto
    next
      show "\<not> \<mu>\<bar>l defines lz" using mt_cons False by (auto simp add: dom_def without_def)
    next
      show "M\<bar>l \<Turnstile> OK" using mt_cons wf_env_subset[of M "M\<bar>l"] by (auto simp add: without_def)
    next
      show "\<mu>\<bar>l \<Turnstile> Mem" using wf_mem_subset[of \<mu> "\<mu>\<bar>l"] mt_cons by (auto simp add: without_def)
    next
      show "\<Gamma>\<bar>x \<sim> {(xz\<rhd>\<tau>z)}" using mt_cons by (auto simp add: without_def env_compat_def)
    next
      have "\<Gamma>,(xz:\<tau>z)\<bar>x \<Turnstile> OK"
        using mt_cons wf_env_subset[of "\<Gamma>,(xz:\<tau>z)" "\<Gamma>,(xz:\<tau>z)\<bar>x"] by (auto simp add: without_def)
      also have "\<Gamma>,(xz:\<tau>z)\<bar>x = (\<Gamma>\<bar>x),(xz:\<tau>z)" using Hxneq by (auto simp add: without_def dom_def)
      finally show "(\<Gamma>\<bar>x),(xz:\<tau>z) \<Turnstile> OK" by auto
    qed
    also have "(\<Gamma>\<bar>x),(xz:\<tau>z) = \<Gamma>,(xz:\<tau>z)\<bar>x" using Hxneq by (auto simp add: without_def dom_def)
    also have "(\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l"
      using False by (auto simp add: without_def)
    also have "(M\<bar>l),(lz:\<tau>z) = M,(lz:\<tau>z)\<bar>l" using False by (auto simp add: without_def)
    finally show ?thesis using Hin by auto
  qed
qed auto

lemma ty_qual_sanity:
  assumes "\<Gamma>;M \<turnstile> v : \<tau>"
  and "value v"
  shows "qualV v = qualT \<tau>"
using prems by (cases, auto)

lemma storeTyping_no_loc:
  assumes "M \<Turnstile> S : \<Gamma>"
  and "\<not> M defines l"
  shows "\<not> l free loc in st S"
using prems by (induct, auto simp add: ty_no_loc free_loc_store_def)

lemma storeTyping_sanity:
  assumes "M \<Turnstile> S : \<Gamma>s"
  and "(x\<mapsto>v) \<in> S"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  shows "qualV v = qualT \<tau> \<and> (\<exists>\<Gamma>v. \<Gamma>v;M \<turnstile> v : \<tau> \<and> un \<Gamma>v \<subseteq> un \<Gamma>s)"
using prems proof induct
  case (st_cons M S \<Gamma>s \<Gamma>c \<sigma> w z)
  thus ?case proof (cases "x=z")
    case True
    have "\<not> (\<Gamma>s\<union>\<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>c" \<Gamma>s] by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>) \<Turnstile> OK" using wf_env_cons by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems by fast
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(x:\<tau>))" by (auto simp add: un_def)
    hence "un \<Gamma>c \<subseteq> un (\<Gamma>s,(x:\<tau>))" using st_cons by auto
    thus ?thesis using prems exI[of _ \<Gamma>c] Hteq Hweq by (auto simp add: ty_qual_sanity)
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_cons by auto
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>\<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    then obtain \<Gamma>v where Htyp: "\<Gamma>v;M \<turnstile> v : \<tau>" and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>c)" using st_cons by auto
    have "un (\<Gamma>s \<union> \<Gamma>c) = un \<Gamma>s \<union> un \<Gamma>c" by (auto simp add: un_def)
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>c" using Hun by auto
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s" using prems by auto
    moreover have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
    ultimately have "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by auto
    moreover have "value v" using prems store_ok_is_value by auto
    ultimately show ?thesis using Htyp ty_qual_sanity by auto
  qed
qed auto

lemma storeTyping_elim:
  assumes "M \<Turnstile> S : \<Gamma>s"
  and "(x\<mapsto>v) \<in> S"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  and "qualT \<tau> = qL"
  shows "\<exists>\<Gamma>v. \<Gamma>v;M \<turnstile> v : \<tau> \<and> M \<Turnstile> S\<bar>x : \<Gamma>s\<bar>x \<union> \<Gamma>v \<and> un \<Gamma>v \<subseteq> un \<Gamma>s \<and>
              \<Gamma>s\<bar>x \<sim> \<Gamma>v \<and> \<Gamma>s\<bar>x \<union> \<Gamma>v \<Turnstile> OK"
using prems proof induct
  case (st_cons M S \<Gamma>s \<Gamma>c \<sigma> w z)
  thus ?case proof (cases "x=z")
    case True
    have "\<not> (\<Gamma>s \<union> \<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence Hznoting: "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s \<union> \<Gamma>c" \<Gamma>s] by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>) \<Turnstile> OK" using wf_env_cons by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems by fast
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(x:\<tau>))" by (auto simp add: un_def)
    hence H: "un \<Gamma>c \<subseteq> un (\<Gamma>s,(x:\<tau>))" using st_cons by auto
    hence "\<Gamma>c;M \<turnstile> v: \<tau> \<and> M \<Turnstile> S : \<Gamma>s \<union> \<Gamma>c \<and> \<Gamma>s \<sim> \<Gamma>c \<and> \<Gamma>s \<union> \<Gamma>c \<Turnstile> OK" 
      using st_cons Hteq Hweq True by auto
    also have "\<Gamma>s = \<Gamma>s,(z:\<sigma>)\<bar>x" using Hznoting True by (auto simp add: without_def dom_def)
    also have "S = S,(z\<mapsto>w)\<bar>x" using st_cons True by (auto simp add: without_def dom_def)
    finally show ?thesis using H exI[of _ \<Gamma>c] Hteq Hweq True by auto
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_cons by auto
    have Hnotin: "\<not> \<Gamma>c defines x" using st_cons Hin compat_ok_sanity[of \<Gamma>s \<Gamma>c x \<tau>] by auto
    hence Hceq: "\<Gamma>c = \<Gamma>c\<bar>x" by (auto simp add: without_def dom_def)
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>\<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    then obtain \<Gamma>v where Htyp: "\<Gamma>v;M \<turnstile> v : \<tau>"
      and Hst: "M \<Turnstile> S\<bar>x : (\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v"
      and Hcompat: "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<sim> \<Gamma>v"
      and Hsubset: "un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>c)"
      and Hok: "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v \<Turnstile> OK" using st_cons by auto
    have "(\<Gamma>s \<union> \<Gamma>c)\<bar>x = \<Gamma>s\<bar>x \<union> \<Gamma>c" using Hnotin by (auto simp add: without_def dom_def)
    hence "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x \<union> \<Gamma>v) \<union> \<Gamma>c" by auto
    hence Hgst': "M \<Turnstile> S\<bar>x : (\<Gamma>s\<bar>x \<union> \<Gamma>v) \<union> \<Gamma>c" using Hst by auto
    have Hnsxz: "\<not> S\<bar>x defines z" using st_cons by (auto simp add: without_def dom_def)
    have "M \<Turnstile> (S\<bar>x),(z\<mapsto>w) : (\<Gamma>s\<bar>x  \<union> \<Gamma>v),(z:\<sigma>)" proof
      have "\<Gamma>s\<bar>x \<sim> \<Gamma>c" using st_cons by (auto simp add: env_compat_def without_def)
      moreover
      have "\<Gamma>v \<sim> \<Gamma>c\<bar>x" using Hcompat by (auto simp add: env_compat_def without_def)
      hence "\<Gamma>v \<sim> \<Gamma>c" using Hceq by auto
      ultimately
      show "\<Gamma>s\<bar>x \<union> \<Gamma>v \<sim> \<Gamma>c" by (auto simp add: env_compat_union_right)
    next
      show "S\<bar>x \<Turnstile> Store" using wf_store_subset[of S "S\<bar>x"] prems by (auto simp add: without_def)
    next
      have "(\<Gamma>s \<union> \<Gamma>c) \<bar> x \<union> \<Gamma>v = \<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c" using Hceq by (auto simp add: without_def)
      thus "\<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c \<Turnstile> OK" using Hok by auto
    next
      have "\<Gamma>s \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s \<union> \<Gamma>c" \<Gamma>s] by auto
      hence "un \<Gamma>s = un (\<Gamma>s\<bar>x)" using prems un_without_l_sanity by auto
      hence "un \<Gamma>c \<subseteq> un (\<Gamma>s\<bar>x) \<union> un \<Gamma>v" using st_cons by auto
      also have "un (\<Gamma>s\<bar>x) \<union> un \<Gamma>v = un (\<Gamma>s\<bar>x \<union> \<Gamma>v)" by (auto simp add: un_def)
      finally show "un \<Gamma>c \<subseteq> un (\<Gamma>s\<bar>x \<union> \<Gamma>v)" by auto
    qed (auto simp add: Hgst' Hnsxz)
    also have "(S\<bar>x),(z\<mapsto>w) =  S,(z\<mapsto>w)\<bar>x" using False by (auto simp add: without_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v" using False by (auto simp add: without_def)
    finally
    have Hgst: "M \<Turnstile> S,(z\<mapsto>w)\<bar>x : ((\<Gamma>s,(z:\<sigma>)\<bar>x) \<union> \<Gamma>v)" by auto
    hence "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c) defines z" using Hgst' Hnsxz store_dom_not_in[of M "S\<bar>x"] by auto
    hence Htemp: "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v) defines z" by (auto simp add: dom_def)
    hence Hznotin: "\<not> \<Gamma>v defines z" by (auto simp add: dom_def)
    have "\<Gamma>s\<bar>x \<sim> \<Gamma>v" using Hcompat by (auto simp add: without_def env_compat_def)
    hence "(\<Gamma>s\<bar>x),(z:\<sigma>) \<sim> \<Gamma>v" using Hznotin by (auto simp add: env_compat_def dom_def)
    also have "(\<Gamma>s\<bar>x),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x" using False by (auto simp add: without_def dom_def)
    finally have Hgcomp: "\<Gamma>s,(z:\<sigma>)\<bar>x \<sim> \<Gamma>v" using False by auto
    have "\<Gamma>s\<bar>x \<union> \<Gamma>v \<Turnstile> OK" using Hok wf_env_subset[of _ "\<Gamma>s\<bar>x \<union> \<Gamma>v"] by (auto simp add: without_def)
    hence "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) \<Turnstile> OK" using Htemp by (auto simp add: wf_env_def dom_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v"
      using False by (auto simp add: without_def dom_def)
    finally have Hgok: "\<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v \<Turnstile> OK" by auto
    have "un (\<Gamma>s \<union> \<Gamma>c) = un \<Gamma>s \<union> un \<Gamma>c" by (auto simp add: un_def)
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>c" using Hsubset by auto
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s" using prems by auto
    moreover have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
    ultimately have "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by auto
    thus ?thesis using Htyp Hgst Hgok Hgcomp by auto
  qed
qed auto

lemma storeTyping_strengthen:
  assumes "M \<Turnstile> S : \<Gamma>"
  and "\<not> l free loc in st S"
  shows "M\<bar>l \<Turnstile> S : \<Gamma>"
using prems proof induct
  case (st_cons M S \<Gamma>s \<Gamma>v \<tau> v x)
  hence "\<not> l free loc in v" by (auto simp add: free_loc_store_def)
  hence "\<Gamma>v; M\<bar>l \<turnstile> v : \<tau>" using st_cons ty_mem_strengthen by auto
  moreover
  have "\<not> l free loc in st S" using st_cons by (auto simp add: free_loc_store_def)
  hence "M\<bar>l \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>v)" using st_cons by auto
  ultimately
  show ?case using st_cons by auto
qed auto

lemma st_wf_store:
  assumes "M \<Turnstile> S : \<Gamma>"
  shows "S \<Turnstile> Store"
using prems proof induct
  case st_nil thus ?case by (auto simp add: wf_store_def)
next
  case st_cons thus ?case by (auto simp add: wf_store_cons)
qed

lemma mt_wf_mem:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  shows "\<mu> \<Turnstile> Mem"
using prems proof induct
  case mt_nil thus ?case using wf_mem_def by auto
next
  case (mt_cons M \<Gamma> \<mu> \<tau> l x) thus ?case using wf_mem_cons[of \<mu> l "TmFreeVar x"] by auto
qed

lemma mt_wf_loc:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  shows "M \<Turnstile> OK"
using prems proof induct
  case (mt_cons M \<Gamma> \<mu> \<tau> l x)
  have "\<not> M defines l" using mt_cons mem_dom_not_in by auto
  thus ?case using mt_cons by (auto simp add: wf_env_cons)
qed (auto simp add: wf_env_def)

lemma storeTyping_wf_env:
  assumes "M \<Turnstile> S : \<Gamma>"
  shows "\<Gamma> \<Turnstile> OK"
using prems proof induct
  case st_nil thus ?case by (auto simp add: wf_env_def)
next
  case (st_cons M S \<Gamma>s \<Gamma>v \<tau> v x)
  have "\<Gamma>s \<Turnstile> OK" using st_cons wf_env_subset[of "\<Gamma>s\<union>\<Gamma>v" \<Gamma>s] by auto
  moreover have "\<not> \<Gamma>s defines x"
    using st_cons store_dom_not_in[of M S _ x] by (auto simp add: dom_def)
  ultimately show ?case using wf_env_cons st_cons by auto
qed

lemma value_preservation:
  assumes "value v"
  and "M \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>m)"
  and Has: "\<Gamma>m \<Turnstile>' \<mu> : M"
  and "\<Gamma>e;{} \<turnstile> v : \<tau>"
  and "\<Gamma>e \<union> \<Gamma>r \<subseteq> \<Gamma>s"
  and "\<Gamma>e \<sim> \<Gamma>r"
  and "\<Gamma>e \<union> \<Gamma>r \<Turnstile> OK"
  and "\<Gamma>s \<sim> \<Gamma>m"
  and "INV2 S"
  and "INV3 S"
  and "\<not> S defines z"
  shows "\<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
          (M' \<Turnstile> S,(z\<mapsto>v) : (\<Gamma>s' \<union> \<Gamma>m')) \<and>
          (\<Gamma>m' \<Turnstile>' \<mu> : M') \<and>
          (\<Gamma>e';{} \<turnstile> TmFreeVar z : \<tau>) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and>
          (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          INV2 (S,(z\<mapsto>v)) \<and> INV3(S,(z\<mapsto>v))"
proof -
  have Hg1: "{z\<rhd>\<tau>};{} \<turnstile> TmFreeVar z : \<tau>" using mt_wf_loc
    by (auto simp add: wf_env_def without_def qualGrelL_def)
  have Hnotin: "\<not> (\<Gamma>s\<union>\<Gamma>m) defines z" using prems store_dom_not_in by auto
  hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
  hence Hgr: "\<not> \<Gamma>r defines z" using prems by (auto simp add: dom_def)
  hence Hg2: "{z\<rhd>\<tau>} \<sim> \<Gamma>r" by (auto simp add: env_compat_def dom_def)
  have Hg4: "INV2 (S,(z\<mapsto>v))" using inv2_cons ty_no_locs prems by auto
  let ?\<Gamma>s' = "(\<Gamma>s - lin \<Gamma>e)"
  have Heq: "?\<Gamma>s' \<union> \<Gamma>e = \<Gamma>s" using prems lin_def by auto
  hence Heqm: "?\<Gamma>s' \<union> \<Gamma>m \<union> \<Gamma>e = \<Gamma>s \<union> \<Gamma>m" using prems lin_def by auto
  have A1: "\<Gamma>e;M \<turnstile> v : \<tau>" using prems mt_wf_loc weakening_m[of \<Gamma>e "{}" v \<tau> M] by auto
  have "\<Gamma>m \<sim> \<Gamma>e" using prems env_compat_subset[of \<Gamma>s \<Gamma>m \<Gamma>e] env_compat_symm by auto
  moreover
  have "\<Gamma>s - lin \<Gamma>e \<sim> \<Gamma>e" using prems proof (auto simp add: env_compat_def lin_def)
    fix x \<tau>
    assume "qualT \<tau> \<noteq> qL"
    thus "qualT \<tau> \<preceq> qU" by (cases "qualT \<tau>", auto)
  qed
  ultimately
  have A2: "(?\<Gamma>s' \<union> \<Gamma>m) \<sim> \<Gamma>e" using prems by (auto simp add: env_compat_def)
  have "un \<Gamma>e \<subseteq> un \<Gamma>s" using prems by (auto simp add: un_def)
  moreover
  have "un (\<Gamma>s - lin \<Gamma>e \<union> \<Gamma>m) = un \<Gamma>s \<union> un \<Gamma>m" by (auto simp add: un_def lin_def)
  ultimately
  have A3:  "un \<Gamma>e \<subseteq> un (\<Gamma>s - lin \<Gamma>e \<union> \<Gamma>m)" by auto
  have "M \<Turnstile> S,(z\<mapsto>v) : (?\<Gamma>s' \<union> \<Gamma>m),(z:\<tau>)" 
    using A1 A2 A3 Heqm st_wf_store  storeTyping_wf_env prems by auto
  hence Hg5: "M \<Turnstile> S,(z\<mapsto>v) : (?\<Gamma>s',(z:\<tau>) \<union> \<Gamma>m)" by auto
  have "\<Gamma>r \<sim> \<Gamma>e" using prems env_compat_symm by auto
  hence "\<Gamma>r \<inter> lin \<Gamma>e = {}" proof (auto simp add: env_compat_def lin_def)
    fix x \<tau>
    assume "\<forall>x \<tau>. (x, Term \<tau>) \<in> \<Gamma>r \<and> (x, Term \<tau>) \<in> \<Gamma>e \<longrightarrow> qualT \<tau> \<preceq> qU"
      and "(x\<rhd>\<tau>) \<in> \<Gamma>r"
      and "(x\<rhd>\<tau>) \<in> \<Gamma>e"
      and "qualT \<tau> = qL"
    hence "qualT \<tau> \<preceq> qU" by force
    thus False using prems by auto
  qed
  hence Hg6: "({z\<rhd>\<tau>} \<union> \<Gamma>r) \<subseteq> ?\<Gamma>s',(z:\<tau>)" using prems by auto
  have "?\<Gamma>s' \<sim> \<Gamma>m" using prems env_compat_subset[of \<Gamma>s \<Gamma>m ?\<Gamma>s'] by auto
  moreover
  have "\<not> \<Gamma>m defines z" 
    using prems store_dom_not_in[of _ _ "\<Gamma>s\<union>\<Gamma>m"] by (auto simp add: dom_def)
  hence "{z\<rhd>\<tau>} \<sim> \<Gamma>m" by (auto simp add: env_compat_def dom_def)
  ultimately
  have Hg7: "?\<Gamma>s',(z:\<tau>) \<sim> \<Gamma>m"
    using env_compat_union_right[of "?\<Gamma>s'" \<Gamma>m "{z\<rhd>\<tau>}"] by auto
  have "\<Gamma>r \<subseteq> \<Gamma>s\<union>\<Gamma>m" using prems by auto
  hence "\<Gamma>r \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>r] storeTyping_wf_env by auto
  hence Hg9: "{z\<rhd>\<tau>} \<union> \<Gamma>r \<Turnstile> OK" using Hgr wf_env_cons by auto
  have Hg10: "INV3 (S,(z\<mapsto>v))" using prems ty_no_locs by auto
  show ?thesis
    apply (rule exI[of _ "{z\<rhd>\<tau>}"], rule exI[of _ "?\<Gamma>s',(z:\<tau>)"])
    apply (rule exI[of _ \<Gamma>m], rule exI[of _ M])
    using Hg1 Hg2 Hg4 Hg5 Hg6 Hg7 Hg9 Hg10 Has by auto
qed

lemma memTyping_update:
  assumes "\<Gamma>m \<Turnstile>' \<mu> : M"
  and "\<mu> defines l"
  and "l\<rhd>\<tau> \<in> M"
  and "\<Gamma>m,(z:\<tau>) \<Turnstile> OK"
  and "\<Gamma>m \<sim> {z\<rhd>\<tau>}"
  shows "\<Gamma>m,(z:\<tau>) \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
using prems proof induct
  case mt_nil thus ?case by (auto simp add: update_def dom_def)
next
  case (mt_cons M \<Gamma> \<mu> \<tau>z lz xz)
  {
    assume Heq: "l = lz"
    have "lz\<rhd>\<tau> \<in> M,(lz:\<tau>z)" using mt_cons Heq by auto
    moreover
    have "\<not> M defines lz" using mt_cons mem_dom_not_in by auto
    hence "M,(lz:\<tau>z) \<Turnstile> OK" using mt_cons wf_env_cons by auto
    ultimately
    have Ht: "\<tau> = \<tau>z" by (auto simp add: wf_env_def)
    hence "\<Gamma>,(xz:\<tau>) \<Turnstile> OK" using prems by auto
    moreover
    have "\<Gamma> \<sim> {(xz\<rhd>\<tau>)}" using prems Ht by auto
    ultimately have "\<Gamma>,(xz:\<tau>) \<Turnstile>' \<mu>:M"
      using prems memTyping_weakening[of \<Gamma> \<mu> M xz \<tau>] by auto
    moreover
    have "\<Gamma>,(xz:\<tau>),(z:\<tau>) \<Turnstile> OK" using Ht prems by auto
    ultimately
    have "(\<Gamma>,(xz:\<tau>)),(z:\<tau>) \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar z) : M,(lz:\<tau>)"
      using prems Ht by (auto simp add: env_compat_def)
    also
    have "\<not> \<mu> defines l" using prems by auto
    hence "\<mu>,(lz\<mapsto>TmFreeVar z) = update (\<mu>,(lz\<mapsto>TmFreeVar xz)) l (TmFreeVar z)"
      using Heq by (auto simp add: update_def dom_def)
    finally have ?case using Ht by auto
  }
  moreover
  {
    assume Hneq: "l \<noteq> lz"
    hence "\<mu> defines l" using prems by (auto simp add: update_def dom_def)
    moreover
    have "(l\<rhd>\<tau>) \<in> M" using prems by auto
    moreover
    have "\<Gamma>,(z:\<tau>) \<Turnstile> OK"
      using prems wf_env_subset[of "\<Gamma>,(xz:\<tau>z),(z:\<tau>)" "\<Gamma>,(z:\<tau>)"] by auto
    moreover
    have "\<Gamma> \<sim> {z\<rhd>\<tau>}" using prems by (auto simp add: env_compat_def)
    ultimately
    have Hmt: "\<Gamma>,(z:\<tau>) \<Turnstile>' update \<mu> l (TmFreeVar z):M" using prems by auto
    moreover
    have "\<Gamma>,(z:\<tau>) \<sim> {xz\<rhd>\<tau>z}" using prems by (auto simp add: env_compat_def)
    moreover
    have Heq: "\<Gamma>,(z:\<tau>),(xz:\<tau>z) = \<Gamma>,(xz:\<tau>z),(z:\<tau>)" by auto
    hence "\<Gamma>,(z:\<tau>),(xz:\<tau>z) \<Turnstile> OK" using prems by auto
    moreover
    have "\<not>  (update \<mu> l (TmFreeVar z)) defines lz"
      using prems Hneq by (auto simp add: update_def dom_def)
    moreover
    have "update \<mu> l (TmFreeVar z) \<Turnstile> Mem" using Hmt mt_wf_mem by auto
    ultimately
    have "\<Gamma>,(z:\<tau>),(xz:\<tau>z) \<Turnstile>' (update \<mu> l (TmFreeVar z)),(lz\<mapsto>TmFreeVar xz) : M,(lz:\<tau>z)"
      using prems by auto
    also note Heq
    also
    have "(update \<mu> l (TmFreeVar z)),(lz\<mapsto>TmFreeVar xz) =
          update (\<mu>,(lz\<mapsto>TmFreeVar xz)) l (TmFreeVar z)"
      using Hneq by (auto simp add: update_def )
    finally have ?case by auto
  }
  ultimately
  show ?case by auto
qed

lemma memTyping_update_qL_without:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "l\<rhd>\<tau> \<in> M"
  and "x\<rhd>\<tau> \<in> \<Gamma>"
  and "qL \<preceq> qualT \<tau>"
  and "\<not> \<Gamma> defines z"
  and "z \<noteq> x"
  shows "\<Gamma>,(z:\<tau>)\<bar>x \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
using prems proof induct
  case mt_nil thus ?case by auto
next
  case (mt_cons M \<Gamma> \<mu> \<tau>a la xa) 
  {
    assume Heq: "la = l"
    have "\<not> M defines la" using mt_cons mem_dom_not_in by auto
    hence Hmok: "M,(la:\<tau>a) \<Turnstile> OK" using wf_env_cons mt_cons by auto
    hence Hteq: "\<tau> = \<tau>a" using mt_cons Heq wf_env_def[of "M,(la:\<tau>a)"] by auto
    have "\<mu>,(l\<mapsto>TmFreeVar xa) \<Turnstile> Mem" using mt_cons wf_mem_cons Heq by auto
    hence Hxeq: "x = xa" 
      using mt_cons Heq wf_mem_def[of "\<mu>,(l\<mapsto>TmFreeVar xa)"] by fast
    have "qualT \<tau> = qL" using mt_cons by (cases "qualT \<tau>", auto)
    hence "\<not> \<Gamma> defines x" using mt_cons Hxeq Hteq compat_ok_sanity_right[of \<Gamma> "{xa\<rhd>\<tau>a}" x \<tau>] by auto
    hence Hgeq: "\<Gamma> = \<Gamma>,(xa:\<tau>)\<bar>x" using Hxeq by (auto simp add: without_def dom_def)
    have "\<Gamma> \<sim> {z\<rhd>\<tau>}" using mt_cons by (auto simp add: env_compat_def dom_def)
    moreover
    have "\<Gamma> \<Turnstile> OK" using mt_cons wf_env_subset[of "\<Gamma>,(xa:\<tau>a)"] by blast
    hence "\<Gamma>,(z:\<tau>) \<Turnstile> OK" using Hteq mt_cons wf_env_cons by (auto simp add: dom_def)
    ultimately
    have "\<Gamma>,(z:\<tau>) \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar z) : M,(l:\<tau>)" using mt_cons Heq by auto
    also note Hgeq
    also have "(\<Gamma>,(xa:\<tau>)\<bar>x),(z:\<tau>) = \<Gamma>,(xa:\<tau>),(z:\<tau>)\<bar>x" 
      using mt_cons by (auto simp add: without_def)
    finally have "\<Gamma>,(xa:\<tau>a),(z:\<tau>)\<bar>x \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar z) : M,(l:\<tau>a)" 
      using Hteq by auto 
    also have "\<mu>,(l\<mapsto>TmFreeVar z) = update (\<mu>,(l\<mapsto>TmFreeVar xa)) l (TmFreeVar z)" 
      using Heq mt_cons by (auto simp add: update_def dom_def)
    finally have ?case using Heq by auto
  }
  moreover
  {
    assume Hneq: "la \<noteq> l"
    have Hzxa: "z \<noteq> xa" using mt_cons by (auto simp add: dom_def)
    have Hznotin: "\<not> (\<Gamma>\<bar>x \<union> {xa\<rhd>\<tau>a}) defines z" 
      using mt_cons by (auto simp add: dom_def without_def)
    have Hxm: "l\<mapsto>TmFreeVar x \<in> \<mu>" using Hneq mt_cons by auto
    moreover
    have Hlm: "l\<rhd>\<tau> \<in> M" using Hneq mt_cons by auto
    moreover
    have Htql: "qualT \<tau> = qL" using mt_cons by (cases "qualT \<tau>", auto)
    have Hgx: "x\<rhd>\<tau> \<in> \<Gamma>" using mt_cons Hlm Hxm memTyping_elim_qL[of \<Gamma> \<mu> M l x \<tau>] by auto
    ultimately
    have "\<Gamma>,(z:\<tau>)\<bar>x \<Turnstile>' update \<mu> l (TmFreeVar z):M" using mt_cons by (auto simp add: dom_def)
    moreover have Hgeq: "(\<Gamma>\<bar>x),(z:\<tau>) = \<Gamma>,(z:\<tau>)\<bar>x" using mt_cons by (auto simp add: without_def)
    moreover have "\<not> (update \<mu> l (TmFreeVar z)) defines la" 
      using Hneq mt_cons by (auto simp add: update_def dom_def)
    moreover have "update \<mu> l (TmFreeVar z) \<Turnstile> Mem" using mt_cons mem_update_ok by auto
    moreover
    have "(\<Gamma>\<bar>x) \<sim> {xa\<rhd>\<tau>a}"
      using mt_cons env_compat_subset[of \<Gamma> _ "\<Gamma>\<bar>x"] by (auto simp add: without_def)
    hence "(\<Gamma>\<bar>x),(z:\<tau>) \<sim> {xa\<rhd>\<tau>a}" using Hzxa by (auto simp add: env_compat_def)
    moreover 
    {
      have "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a} \<Turnstile> OK" 
        using mt_cons wf_env_subset[of "\<Gamma>,(xa:\<tau>a)" "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}"] 
        by (auto simp add: without_def)
      hence "((\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}),(z:\<tau>) \<Turnstile> OK" 
        using mt_cons Hzxa wf_env_cons[of "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}" z \<tau>] Hznotin by auto
      also have "((\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}),(z:\<tau>) = (\<Gamma>\<bar>x),(z:\<tau>) \<union> {xa\<rhd>\<tau>a}" by auto
      finally have "(\<Gamma>\<bar>x),(z:\<tau>) \<union> {xa\<rhd>\<tau>a} \<Turnstile> OK" by auto
    }
    ultimately
    have "(\<Gamma>\<bar>x),(z:\<tau>),(xa:\<tau>a) \<Turnstile>' (update \<mu> l (TmFreeVar z)),(la\<mapsto>TmFreeVar xa) : M,(la:\<tau>a)" 
      using mt_cons by auto
    also note Hgeq
    also 
    have "x \<noteq> xa" 
      using mt_cons Htql Hgx compat_ok_sanity[of \<Gamma> "{xa\<rhd>\<tau>a}" x \<tau>] by (auto simp add: dom_def)
    hence "(\<Gamma>,(z:\<tau>)\<bar>x),(xa:\<tau>a) = \<Gamma>,(xa:\<tau>a),(z:\<tau>)\<bar>x" by (auto simp add: without_def)
    also have "(update \<mu> l (TmFreeVar z)),(la\<mapsto>TmFreeVar xa) =
               update (\<mu>,(la\<mapsto>TmFreeVar xa)) l (TmFreeVar z)" 
      using Hneq by (auto simp add: update_def)
    finally have ?case by auto
  }
  ultimately show ?case by auto
qed

lemma storeTyping_wf_store:
  assumes "M \<Turnstile> S : \<Gamma>"
  shows "S \<Turnstile> Store"
using prems proof induct
  case st_nil thus ?case by (auto simp add: wf_store_def)
next
  case (st_cons M S \<Gamma>s \<Gamma>v \<tau> v x)
  thus ?case using wf_store_cons by auto
qed

lemma memTyping_wf_mem:
  assumes "\<Gamma> \<Turnstile>' \<mu> : M"
  shows "\<mu> \<Turnstile> Mem"
using prems proof induct
  case mt_nil thus ?case by (auto simp add: wf_mem_def)
next
  case (mt_cons M \<Gamma> \<mu> \<tau> l x)
  thus ?case using wf_mem_cons by auto
qed

theorem preservation:
  assumes "\<Gamma>e;Mz \<turnstile> e : \<tau>"
  and "Mz = {}"
  and "M \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>m)"
  and "\<Gamma>m \<Turnstile>' \<mu> : M"
  and "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'"
  and "\<Gamma>e \<union> \<Gamma>r \<subseteq> \<Gamma>s"
  and "\<Gamma>e \<sim> \<Gamma>r"
  and "\<Gamma>e \<union> \<Gamma>r \<Turnstile> OK"
  and "\<Gamma>s \<sim> \<Gamma>m"
  and "INV2 S"
  and "INV3 S"
  shows "\<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
          (M' \<Turnstile> S' : (\<Gamma>s' \<union> \<Gamma>m')) \<and>
          (\<Gamma>m' \<Turnstile>' \<mu>' : M') \<and>
          (\<Gamma>e';{} \<turnstile> e' : \<tau>) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and>
          (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          INV2 S' \<and> INV3 S'"
proof -
have "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" using prems storeTyping_wf_env by auto
from prems this show ?thesis proof (induct arbitrary: e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  case (t_unit M \<Gamma>e q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m)
  have "S;\<mu>;TmUnit q \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    note value_preservation[of "TmUnit q" M S \<Gamma>s \<Gamma>m \<mu> \<Gamma>e "Qual q TyUnit" \<Gamma>r z]
    also have "S,(z\<mapsto>TmUnit q)=S'" using e_val by auto
    finally show ?thesis using prems by auto
  qed auto
next
  case (t_loc M \<Gamma>e \<tau> l q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m)
  have "S;\<mu>;TmLoc q l \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    note value_preservation[of "TmLoc q l" M S \<Gamma>s \<Gamma>m \<mu>
                               \<Gamma>e " Qual q (TyRef \<tau>)" \<Gamma>r z]
    also have "S,(z\<mapsto>TmLoc q l)=S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using prems by auto
  qed auto
next
  case (t_abs L M \<Gamma>e \<tau>1 \<tau>2 e q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m)
  have "S;\<mu>;\<lambda>^q[\<tau>1].e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    note value_preservation[of "\<lambda>^q[\<tau>1].e" M S \<Gamma>s \<Gamma>m \<mu> \<Gamma>e "Qual q (\<tau>1\<rightarrow>\<tau>2)" \<Gamma>r z]
    also have "S,(z\<mapsto>\<lambda>^q[\<tau>1].e)=S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using t_abs e_val by auto
  qed auto
next
  case (t_var M \<Gamma> \<tau> x e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m)
  have "S;\<mu>;TmFreeVar x \<hookrightarrow> S';\<mu>';e'" .
  thus ?case by (cases, auto)
next
  case (t_app Mz \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmApp e1 e2 \<hookrightarrow> S';\<mu>';e'"  .
  thus ?case proof cases
    case (e_app_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a)
    from t_app
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
           (\<Gamma>e';{} \<turnstile> e1' : Qual q (\<tau>1 \<rightarrow> \<tau>2)) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_app_p1 by auto
    moreover
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using t_app env_compat_union_left by auto
    moreover
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK" using t_app by auto
    moreover
    have "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_app by auto
    moreover
    have "INV3 S" by assumption
    ultimately
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';{} \<turnstile> e1' : Qual q (\<tau>1 \<rightarrow> \<tau>2)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2"] by auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;{} \<turnstile> e1' \<cdot> e2 : \<tau>2" proof
      show "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1' \<union> \<Gamma>2 \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" "\<Gamma>1' \<union> \<Gamma>2"] by auto
    next
      show "\<Gamma>1';{} \<turnstile> e1' : Qual q (\<tau>1 \<rightarrow> \<tau>2)" using Ht .
    next
      show "\<Gamma>2;{} \<turnstile> e2 : \<tau>1" using t_app by auto
    qed
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_app env_compat_def by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_app_p1 by auto
  next
    case (e_app_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a) 
    from t_app
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
         \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and> 
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';{} \<turnstile> e2' : \<tau>1) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_app_p2 by auto
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_app env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    moreover
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK" using t_app by auto
    moreover
    have "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_app by auto
    moreover
    have "INV3 S" by assumption
    ultimately 
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';{} \<turnstile> e2' : \<tau>1"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1"] by auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';{} \<turnstile> e1 \<cdot> e2' : \<tau>2" proof
      show "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1 \<union> \<Gamma>2' \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" "\<Gamma>1 \<union> \<Gamma>2'"] by auto
    next
      show "\<Gamma>1;{} \<turnstile> e1 : Qual q  (\<tau>1 \<rightarrow> \<tau>2)" using t_app by auto
    next 
      show "\<Gamma>2';{} \<turnstile> e2' : \<tau>1" using Ht by auto 
    qed
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_app env_compat_def by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto 
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M']) 
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_app_p2 by auto
  next
    case (e_beta Sa Sa' \<mu> \<tau> e qz y z)
    {
      assume Hu: "qz = qU"
      hence Heq: "S = S'" and Hzins: "z\<mapsto>\<lambda>^qz[\<tau>].e \<in> S" using e_beta t_app by auto
      have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" using e_beta t_app by auto
      also have Hgmeq: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s\<union>(un \<Gamma>m) \<union> \<Gamma>m" by (auto simp add: un_def)
      finally have Hg1: "M \<Turnstile> S:((\<Gamma>s\<union>(un \<Gamma>m)) \<union> \<Gamma>m)" by auto
      have  Hg2: "\<Gamma>m \<Turnstile>' \<mu>:M"  using e_beta t_app by auto
      have Hzing: "z\<rhd>(Qual q (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>1" using t_app e_beta by auto
      hence Hzing: "z\<rhd>(Qual q (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m" using t_app e_beta by auto
      have Hyg2: "y\<rhd>\<tau>1 \<in> \<Gamma>2" using t_app e_beta by auto
      hence Hying: "y\<rhd>\<tau>1 \<in> \<Gamma>s" using t_app e_beta by auto
      have Htr: "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
      have "\<exists>\<Gamma>v. \<Gamma>v;M \<turnstile> \<lambda>^qz[\<tau>].e : Qual q (\<tau>1 \<rightarrow> \<tau>2) \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz[\<tau>].e"
                                 "Qual q (\<tau>1 \<rightarrow> \<tau>2)"] Hzins Hzing Htr
        by auto
      then obtain \<Gamma>v
        where Ht: "\<Gamma>v;M \<turnstile> \<lambda>^qz[\<tau>].e : Qual q (\<tau>1 \<rightarrow> \<tau>2)"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        by force
      hence Hqeq: "qz = q" and Hteq: "\<tau> = \<tau>1" by auto
      hence HqualG: "qualG \<Gamma>v \<preceq> qU" using Hu Ht by auto
      hence Hunv: "un \<Gamma>v = \<Gamma>v" using equal_unrestricted by auto
      have "\<Gamma>v,(y:\<tau>1);M \<turnstile> freshen_tm y e : \<tau>2" using Ht proof
        fix L \<tau>2a
        assume Ht: "\<forall>x. x \<notin> set (free_tm e) \<and> x \<notin> L \<and> x \<notin> Environ.dom \<Gamma>v \<longrightarrow>
                \<Gamma>v,(x:\<tau>);M \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2a"
          and Hq: "Qual q (\<tau>1 \<rightarrow> \<tau>2) = Qual qz (\<tau> \<rightarrow> \<tau>2a)"
          and Hlfin: "finite L"
        moreover
        have "finite (dom \<Gamma>v)"
          using ty_wf_env[of \<Gamma>v] wf_env_def[of \<Gamma>v] dom_finite[of \<Gamma>v] prems by auto
        then obtain w where "\<not> \<Gamma>v defines w" and "\<not> w free in e" and "w \<notin> L" and "w\<noteq>y" 
          using Hlfin fresh_exists[of "dom \<Gamma>v \<union> set (free_tm e) \<union> L \<union> {y}"] by auto
        moreover
        have "\<tau>2 = \<tau>2a" using Hq by auto
        ultimately have "\<Gamma>v,(w:\<tau>1);M \<turnstile> freshen_tm w e : \<tau>2" by auto
        moreover
        have "\<Gamma>s \<union> \<Gamma>v \<subseteq> \<Gamma>s \<union> \<Gamma>m" using Hunv Hun by (auto simp add: un_def)
        hence "\<Gamma>s \<union> \<Gamma>v \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" "\<Gamma>s\<union>\<Gamma>v"] t_app by auto
        hence "\<Gamma>v,(y:\<tau>1) \<Turnstile> OK" using wf_env_union_sanity_cons Hying by auto
        moreover have "\<Gamma>v \<sim> {(y\<rhd>\<tau>1)}" using HqualG 
          by (cases "qualT \<tau>1", 
              auto simp add: env_compat_def dom_def qualGrelL_def qualT_unrestricted)
        ultimately show ?thesis using prems substitution[of e _ w \<tau>1 \<Gamma>v M 0 \<tau>2 y] by auto
      qed
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q l. (\<lambda>^qz[\<tau>].e) = TmLoc q l) \<or> (\<forall>l. \<not> l free loc in (\<lambda>^qz[\<tau>].e))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<lambda>^qz[\<tau>].e))" by auto
      hence "(\<forall>l. \<not> l free loc in e)" by auto
      hence "(\<forall>l. \<not> l free loc in (freshen_tm y e))" using vsubst_tm_same_loc by auto
      ultimately
      have Hg3: "\<Gamma>v,(y:\<tau>1);{} \<turnstile> freshen_tm y e : \<tau>2"
        using ty_mem_strengthen_all[of "\<Gamma>v,(y:\<tau>1)" _ "freshen_tm y e"] by auto
      note Hun
      also have "un (\<Gamma>s \<union> \<Gamma>m) = un \<Gamma>s \<union> un \<Gamma>m" by (auto simp add: un_def)
      also note Hunv
      finally have "\<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>m" by auto
      moreover have "un \<Gamma>s \<subseteq> \<Gamma>s" by (auto simp add: un_def)
      ultimately have "\<Gamma>v \<subseteq> \<Gamma>s \<union> un \<Gamma>m" by auto
      moreover
      note Hying
      moreover have "\<Gamma>r \<subseteq> \<Gamma>s" using t_app by auto
      ultimately have Hg4: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<subseteq> \<Gamma>s\<union>(un \<Gamma>m)" by auto
      have Hg5: "\<Gamma>v,(y:\<tau>1) \<sim> \<Gamma>r" using HqualG proof (auto simp add: env_compat_def qualGrelL_def)
        assume Hy: "y\<rhd>\<tau>1 \<in> \<Gamma>r"
        { assume "qualT \<tau>1 = qU"  hence "qualT \<tau>1 \<preceq> qU" by auto }
        moreover
        {
          assume "qualT \<tau>1 = qL"
          moreover have "\<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK"
            using t_app wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" "\<Gamma>2 \<union> \<Gamma>r"] by auto
          moreover have "\<Gamma>2 \<sim> \<Gamma>r"
            using t_app env_compat_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>r \<Gamma>2] by auto
          ultimately have False using Hy Hyg2 compat_ok_sanity[of \<Gamma>2 \<Gamma>r y \<tau>1]
            by (auto simp add: dom_def)
        }
        ultimately show "qualT \<tau>1 \<preceq> qU" by (cases "qualT \<tau>1", auto)
      qed
      have "\<Gamma>s \<union> (un \<Gamma>m) \<Turnstile> OK"
        using t_app wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" "\<Gamma>s \<union> (un \<Gamma>m)"]
        by (auto simp add: un_def)
      hence Hg6: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s \<union> (un \<Gamma>m)"] Hg4 by auto
      have "qualG (un \<Gamma>m) \<preceq> qU" by (auto simp add: un_def qualGrelL_def)
      hence Hg7: "(\<Gamma>s\<union>(un \<Gamma>m)) \<sim> \<Gamma>m"
        using t_app env_compat_weakening_left[of \<Gamma>s \<Gamma>m "un \<Gamma>m"] by auto
      have Hg10: "INV2 S" using e_beta t_app by auto
      have Hg11: "INV3 S" using e_beta t_app by auto
      have ?thesis
        apply (rule exI[of _ "\<Gamma>v,(y:\<tau>1)"],
          rule exI[of _ "\<Gamma>s\<union> (un \<Gamma>m)"],
          rule exI[of _ \<Gamma>m], rule exI[of _ M])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_beta Heq by auto
    }
    moreover
    {
      assume Hql: "qz = qL"
      hence Heq: "S' = S\<bar>z" and Hzins: "z\<mapsto>\<lambda>^qz[\<tau>].e \<in> S"
        using e_beta t_app by auto
      have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" using e_beta t_app by auto
      have Hzing: "z\<rhd>(Qual q (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>1" using t_app e_beta by auto
      hence Hzing: "z\<rhd>(Qual q (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m" using t_app e_beta by auto
      have Hyg2: "y\<rhd>\<tau>1 \<in> \<Gamma>2" using t_app e_beta by auto
      hence Hying: "y\<rhd>\<tau>1 \<in> \<Gamma>s" using t_app e_beta by auto
      hence "y\<rhd>\<tau>1 \<in> \<Gamma>s\<union>\<Gamma>m" and "\<Gamma>s\<union>\<Gamma>m \<Turnstile> OK" by auto
      hence "y \<noteq> z" using Hzing proof (cases "y=z")
        assume "y = z"
          and "y\<rhd>\<tau>1 \<in> \<Gamma>s\<union>\<Gamma>m"
          and "\<Gamma>s\<union>\<Gamma>m \<Turnstile> OK" 
          and "z\<rhd>(Qual q (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m" 
        hence "\<tau>1 = Qual q (\<tau>1\<rightarrow>\<tau>2)" by (auto simp add: wf_env_def[of "\<Gamma>s\<union>\<Gamma>m"])
        hence "size \<tau>1 = size (Qual q (\<tau>1\<rightarrow>\<tau>2))" by auto
        thus ?thesis by auto
      next
        assume "y\<noteq>z"
        thus ?thesis by auto
      qed
      hence Hyingwz: "y\<rhd>\<tau>1 \<in> \<Gamma>s\<bar>z" using Hying by (auto simp add: without_def)
      have Htr: "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
      have "q = qL"
        using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz[\<tau>].e"
                                 "Qual q (\<tau>1 \<rightarrow> \<tau>2)"] Hzins Hzing Htr Hql
        by auto
      hence "\<exists>\<Gamma>v. \<Gamma>v;M \<turnstile> \<lambda>^qz[\<tau>].e : Qual q (\<tau>1 \<rightarrow> \<tau>2) \<and>
             M \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v) \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>m) \<and>
             ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v) \<and> ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v) \<Turnstile> OK"
        using storeTyping_elim[of M S "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz[\<tau>].e"
                               "Qual q (\<tau>1 \<rightarrow> \<tau>2)"] Hzins Hzing Htr
        by auto
      then obtain \<Gamma>v
        where Ht: "\<Gamma>v;M \<turnstile> \<lambda>^qz[\<tau>].e : Qual q (\<tau>1 \<rightarrow> \<tau>2)"
        and Hst: "M \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v)"
        and Hcomp: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v)"
        and Hok: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v) \<Turnstile> OK"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        by force
      hence Hqeq: "qz = q" and Hteq: "\<tau> = \<tau>1" by auto
      have Hse: "z\<rhd>(Qual qL (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>1" using t_app e_beta Hqeq Hql by auto
      hence Hsz: "z\<rhd>(Qual qL (\<tau>1\<rightarrow>\<tau>2)) \<in> \<Gamma>s" using t_app by auto
      hence "\<not> \<Gamma>m defines z"
        using compat_ok_sanity[of \<Gamma>s \<Gamma>m z "Qual qL (\<tau>1\<rightarrow>\<tau>2)"]  t_app Hql Hqeq
        by auto
      hence Hmz: "\<Gamma>m\<bar>z = \<Gamma>m" by (auto simp add: without_def dom_def)
      note Hst
      also have "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v = \<Gamma>s\<bar>z \<union> \<Gamma>v \<union> \<Gamma>m" using Hmz
        by (auto simp add: without_def)
      finally
      have Hg1: "M \<Turnstile> S\<bar>z:((\<Gamma>s\<bar>z \<union> \<Gamma>v) \<union> \<Gamma>m)" using Hst by auto
      have Hg2: "\<Gamma>m \<Turnstile>' \<mu>:M"  using e_beta t_app by auto
      have "\<Gamma>v,(y:\<tau>1);M \<turnstile> freshen_tm y e : \<tau>2" using Ht proof
        fix L \<tau>2a
        assume Ht: "\<forall>x. x \<notin> set (free_tm e) \<and> x \<notin> L \<and> x \<notin> Environ.dom \<Gamma>v \<longrightarrow>
                \<Gamma>v,(x:\<tau>);M \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2a"
          and Hq: "Qual q (\<tau>1 \<rightarrow> \<tau>2) = Qual qz (\<tau> \<rightarrow> \<tau>2a)"
          and Hlfin: "finite L"
        moreover
        have "finite (dom \<Gamma>v)"
          using ty_wf_env[of \<Gamma>v] wf_env_def[of \<Gamma>v] dom_finite[of \<Gamma>v] prems by auto
        then obtain w where "\<not> \<Gamma>v defines w" and "\<not> w free in e" and "w \<notin> L" and "w\<noteq>y" 
          using Hlfin fresh_exists[of "dom \<Gamma>v \<union> set (free_tm e) \<union> L \<union> {y}"] by auto
        moreover
        have "\<tau>2 = \<tau>2a" using Hq by auto
        ultimately have "\<Gamma>v,(w:\<tau>1);M \<turnstile> freshen_tm w e : \<tau>2" by auto
        moreover
        have "\<Gamma>s\<bar>z \<union> \<Gamma>v \<subseteq> (\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v" by (auto simp add: without_def)
        hence "\<Gamma>s\<bar>z \<union> \<Gamma>v \<Turnstile> OK" using wf_env_subset Hok by auto
        hence "\<Gamma>v,(y:\<tau>1) \<Turnstile> OK" using wf_env_union_sanity_cons Hyingwz by auto
        moreover 
        have "y\<rhd>\<tau>1 \<in> ((\<Gamma>s\<union>\<Gamma>m)\<bar>z)" using Hyingwz by (auto simp add: without_def)
        hence "{(y\<rhd>\<tau>1)} \<sim> \<Gamma>v" using env_compat_subset Hcomp by auto
        hence "\<Gamma>v \<sim> {(y\<rhd>\<tau>1)}" using env_compat_symm by auto
        ultimately show ?thesis using prems substitution[of e _ w \<tau>1 \<Gamma>v M 0 \<tau>2 y] by auto
      qed
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q l. (\<lambda>^qz[\<tau>].e) = TmLoc q l) \<or> (\<forall>l. \<not> l free loc in (\<lambda>^qz[\<tau>].e))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<lambda>^qz[\<tau>].e))" by auto
      hence "(\<forall>l. \<not> l free loc in e)" by auto
      hence "(\<forall>l. \<not> l free loc in (freshen_tm y e))" using vsubst_tm_same_loc by auto
      ultimately
      have Hg3: "\<Gamma>v,(y:\<tau>1);{} \<turnstile> freshen_tm y e : \<tau>2"
        using ty_mem_strengthen_all[of "\<Gamma>v,(y:\<tau>1)" _ "freshen_tm y e"] by auto
      have "\<Gamma>s \<Turnstile> OK" using t_app wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>s] by auto
      moreover
      have "\<tau>1 \<noteq> Qual qL (\<tau>1\<rightarrow>\<tau>2)" proof
        assume "\<tau>1 = Qual qL (\<tau>1 \<rightarrow> \<tau>2)"
        hence "size \<tau>1 = size (Qual qL (\<tau>1 \<rightarrow> \<tau>2))" by auto
        thus False by auto
      qed
      ultimately have Hneq: "y \<noteq> z" using Hying Hsz
        by (auto simp add: wf_env_def)
      have "\<not> \<Gamma>r defines z"
        using compat_ok_sanity[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r z "Qual qL (\<tau>1\<rightarrow>\<tau>2)"] Hql
          Hqeq Hse t_app by auto
      moreover have "\<Gamma>r \<subseteq> \<Gamma>s" using t_app by auto
      ultimately have Hrsz: "\<Gamma>r \<subseteq> \<Gamma>s\<bar>z" by (auto simp add: dom_def without_def)
      moreover
      hence "y\<rhd>\<tau>1 \<in> (\<Gamma>s\<bar>z)" using Hying Hneq by (auto simp add: without_def)
      ultimately have Hg4: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<subseteq> \<Gamma>s\<bar>z \<union> \<Gamma>v" by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>v" using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>r \<sim> \<Gamma>v" using Hrsz env_compat_subset[of "\<Gamma>s\<bar>z" \<Gamma>v \<Gamma>r] by auto
      hence "\<Gamma>v \<sim> \<Gamma>r" by (auto simp add: env_compat_def)
      hence Hg5: "\<Gamma>v,(y:\<tau>1) \<sim> \<Gamma>r"
      proof (auto simp add: env_compat_def qualGrelL_def)
        assume Hy: "y\<rhd>\<tau>1 \<in> \<Gamma>r"
        { assume "qualT \<tau>1 = qU"  hence "qualT \<tau>1 \<preceq> qU" by auto }
        moreover
        {
          assume "qualT \<tau>1 = qL"
          moreover have "\<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK"
            using t_app wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" "\<Gamma>2 \<union> \<Gamma>r"] by auto
          moreover have "\<Gamma>2 \<sim> \<Gamma>r"
            using t_app env_compat_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>r \<Gamma>2] by auto
          ultimately have False using Hy Hyg2 compat_ok_sanity[of \<Gamma>2 \<Gamma>r y \<tau>1]
            by (auto simp add: dom_def)
        }
        ultimately show "qualT \<tau>1 \<preceq> qU" by (cases "qualT \<tau>1", auto)
      qed
      have "\<Gamma>s\<bar>z \<union> \<Gamma>v \<Turnstile> OK"
        using Hok wf_env_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v" "\<Gamma>s\<bar>z \<union> \<Gamma>v"]
        by (auto simp add: without_def)
      hence Hg6: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s\<bar>z \<union> \<Gamma>v"] Hg4 by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>m" using t_app env_compat_subset[of \<Gamma>s \<Gamma>m "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      moreover
      have "\<Gamma>m\<bar>z \<sim> \<Gamma>v" using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>m\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>m \<sim> \<Gamma>v" using Hmz by auto
      hence "\<Gamma>v \<sim> \<Gamma>m" by (auto simp add: env_compat_def)
      ultimately have Hg7: "(\<Gamma>s\<bar>z \<union>\<Gamma>v) \<sim> \<Gamma>m" using env_compat_union_right by auto
      have Hs: "S\<bar>z \<subseteq> S" by (auto simp add: without_def)
      moreover
      have "INV2 S" by assumption
      ultimately
      have Hg10: "INV2 (S\<bar>z)" using inv2_subset by auto
      have "INV3 S" by assumption
      hence Hg11: "INV3 (S\<bar>z)" using Hs inv3_subset by auto
      have ?thesis
        apply (rule exI[of _ "\<Gamma>v,(y:\<tau>1)"],
          rule exI[of _ "\<Gamma>s\<bar>z \<union> \<Gamma>v"],
          rule exI[of _ \<Gamma>m], rule exI[of _ M])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_beta Heq by auto
      hence ?thesis by auto
    }
    ultimately
    show ?thesis by (cases qz, auto)
  qed auto
next
  case (t_new Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_new_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_new
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
           (\<Gamma>e';{} \<turnstile> e' : \<tau>) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (INV2 S') \<and> (INV3 S')" by auto
    have "Mz = {}" and "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'"
      and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
      and "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "INV3 S"
      using prems by auto
    then obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';{} \<turnstile> e' : \<tau>"
      and H4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and H5: "\<Gamma>e' \<sim> \<Gamma>r"
      and H6: "\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'" using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r] by auto
    moreover
    have Hg3: "\<Gamma>e';{} \<turnstile> TmNew e' : Qual qL (TyRef \<tau>)" using Ht by auto
    ultimately show ?thesis using e_new_p by auto
  next
    case (e_new_c Sa \<mu>a l x z)
    have Hm: "M \<Turnstile> OK" using prems mt_wf_loc by auto
    moreover
    have Hlnotinm: "\<not> M defines l" using prems mem_dom_not_in by auto
    ultimately
    have Hmok: "M,(l:\<tau>) \<Turnstile> OK" using e_new_c wf_env_cons by auto
    have "{};M,(l:\<tau>) \<turnstile> TmLoc qL l : Qual qL (TyRef \<tau>)"
      using Hmok by (auto simp add: wf_env_def qualGrelL_def)
    moreover
    have "M,(l:\<tau>) \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m \<union> {})"
      using prems storeTyping_weakening Hmok by auto
    moreover
    have "\<not> S defines z" using prems by auto
    ultimately
    have Hg1: "M,(l:\<tau>) \<Turnstile> S,(z\<mapsto>TmLoc qL l) : (\<Gamma>s \<union> \<Gamma>m),(z:Qual qL (TyRef \<tau>))"
      by (auto simp add: st_cons un_def env_compat_def st_wf_store)
    have Hg3: "{z\<rhd>Qual qL (TyRef \<tau>)}; {} \<turnstile> TmFreeVar z : Qual qL (TyRef \<tau>)"
      by (auto simp add: without_def wf_env_def qualGrelL_def)
    have Hsmnotin: "\<not> (\<Gamma>s\<union>\<Gamma>m) defines z" using prems store_dom_not_in by auto
    hence Hnotin: "\<not> \<Gamma>r defines z" using prems by (auto simp add: dom_def)
    hence Hg5: "{z\<rhd>Qual qL (TyRef \<tau>)} \<sim> \<Gamma>r"
      by (auto simp add: env_compat_def dom_def)
    have "\<Gamma>r \<subseteq> \<Gamma>s\<union>\<Gamma>m" using prems by auto
    hence "\<Gamma>r \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>r] prems by auto
    hence Hg6: "{z\<rhd>Qual qL (TyRef \<tau>)} \<union> \<Gamma>r \<Turnstile> OK" using Hnotin wf_env_cons[of \<Gamma>r z] by auto
    have "\<not> l free loc in st S" using t_new Hlnotinm storeTyping_no_loc by auto
    hence Hg10: "INV2 (S,(z\<mapsto>TmLoc qL l))" using t_new e_new_c inv2_cons_loc[of S z l] by auto
    have "INV3 S" by assumption
    hence Hg11: "INV3 (S,(z\<mapsto>TmLoc qL l))" by auto
    have "\<Gamma>;{} \<turnstile> TmFreeVar x : \<tau>" using prems by auto
    hence Hxin: "(x\<rhd>\<tau>) \<in> \<Gamma>"  by auto
    hence "{x\<rhd>\<tau>} \<subseteq> \<Gamma>"  by auto
    hence Hxsub: "{x\<rhd>\<tau>} \<subseteq> \<Gamma>s" using t_new by auto
    have Hsmcomp: "\<Gamma>s \<sim> \<Gamma>m" and Hsmok: "\<Gamma>s \<union>  \<Gamma>m \<Turnstile> OK" by assumption
    show ?thesis proof (cases "qualT \<tau>")
      case qL
      note Hg1
      also have "\<Gamma>s \<union> \<Gamma>m = (\<Gamma>s - {x\<rhd>\<tau>}) \<union> (\<Gamma>m,(x:\<tau>))" using Hxsub by auto
      also have "(((\<Gamma>s - {x\<rhd>\<tau>})) \<union> (\<Gamma>m,(x:\<tau>))),(z:Qual qL (TyRef \<tau>)) =
                (((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL (TyRef \<tau>))) \<union> (\<Gamma>m,(x:\<tau>)))" by auto
      finally have Hg1: "M,(l:\<tau>) \<Turnstile> S,(z\<mapsto>TmLoc qL l) :
        (((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL (TyRef \<tau>))) \<union> (\<Gamma>m,(x:\<tau>)))"
        by auto
      have Hmx: "\<not> \<Gamma>m defines x"
        using qL Hxsub Hsmcomp Hsmok compat_ok_sanity[of \<Gamma>s \<Gamma>m x \<tau>] by auto
      have Hg2: "\<Gamma>m,(x:\<tau>) \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x):M,(l:\<tau>)" proof
        show "\<Gamma>m \<Turnstile>' \<mu>:M" using t_new by auto
      next
        show "\<Gamma>m \<sim> {(x, Term \<tau>)}"
          using Hmx by (auto simp add: env_compat_def dom_def)
      next
        have "\<Gamma>m \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>m] by auto
        thus "\<Gamma>m,(x:\<tau>) \<Turnstile> OK" using Hmx wf_env_cons by auto
      next
        show "l \<notin> Environ.dom \<mu>" using prems by auto
      next
        show "M \<Turnstile> OK" using Hm by auto
      next
        show "\<mu> \<Turnstile> Mem" using prems mt_wf_mem by auto
      qed 
      have "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r \<Turnstile> OK" by assumption
      hence "\<not> \<Gamma>r defines x"
        using qL Hxin compat_ok_sanity[of \<Gamma> \<Gamma>r x \<tau>] by auto
      moreover
      have "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
      ultimately
      have "\<Gamma>r \<subseteq> \<Gamma>s - {x\<rhd>\<tau>}" by (auto simp add: dom_def)
      hence Hg4: "{(z, Term (Qual qL (TyRef \<tau>)))} \<union> \<Gamma>r \<subseteq>
                 (\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL (TyRef \<tau>))" by auto
      have "\<Gamma>s \<sim> \<Gamma>m" by assumption
      hence Hg7: "((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL (TyRef \<tau>))) \<sim> (\<Gamma>m,(x:\<tau>))"
      using Hsmnotin proof (auto simp add: env_compat_def dom_def)
        assume "Qual qL (TyRef \<tau>) = \<tau>"
        hence "size (Qual qL (TyRef \<tau>)) = size \<tau>" by auto
        thus "qualT \<tau> \<preceq> qU" by auto
      qed
      show ?thesis
        apply (rule exI[of _ "{z\<rhd>Qual qL (TyRef \<tau>)}"],
               rule exI[of _ "(\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL (TyRef \<tau>))"],
               rule exI[of _ "\<Gamma>m,(x:\<tau>)"], rule exI[of _ "M,(l:\<tau>)"])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_new_c by auto
    next
      case qU
      note Hg1
      also have Hueq: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> (\<Gamma>m,(x:\<tau>))" using Hxsub by auto
      also have "(\<Gamma>s \<union> (\<Gamma>m,(x:\<tau>))),(z:Qual qL (TyRef \<tau>)) =
                \<Gamma>s,(z:Qual qL (TyRef \<tau>)) \<union> (\<Gamma>m,(x:\<tau>))" by auto
      finally have Hg1: "M,(l:\<tau>) \<Turnstile> S,(z\<mapsto>TmLoc qL l) :
        (\<Gamma>s,(z:Qual qL (TyRef \<tau>)) \<union> (\<Gamma>m,(x:\<tau>)))"
        by auto
      have Hg2: "\<Gamma>m,(x:\<tau>) \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x):M,(l:\<tau>)" proof
        show "\<Gamma>m \<Turnstile>' \<mu>:M" using t_new by auto
      next
        show "\<Gamma>m \<sim> {(x, Term \<tau>)}"
          using qU by (auto simp add: env_compat_def dom_def)
      next
        have Hgmok: "\<Gamma>m \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>m] by auto
        show "\<Gamma>m,(x:\<tau>) \<Turnstile> OK" proof (cases "\<Gamma>m defines x")
          case True
          have "x\<rhd>\<tau> \<in> \<Gamma>s" using Hxsub by auto
          moreover
          have "\<Gamma>m \<union> \<Gamma>s \<Turnstile> OK" using Hsmok by (auto simp add: wf_env_def)
          ultimately
          have "x\<rhd>\<tau> \<in> \<Gamma>m" using wf_env_union_sanity[of \<Gamma>m \<Gamma>s x \<tau>]
            using True by auto
          hence "\<Gamma>m,(x:\<tau>) = \<Gamma>m" by auto
          thus ?thesis using Hgmok by auto
        next
          case False
          thus ?thesis using Hgmok wf_env_cons by auto
        qed
      next
        show "l \<notin> Environ.dom \<mu>" using prems by auto
      next
        show "M \<Turnstile> OK" using Hm by auto
      next
        show "\<mu> \<Turnstile> Mem" using prems mt_wf_mem by auto
      qed
      have "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
      hence Hg4: "{(z, Term (Qual qL (TyRef \<tau>)))} \<union> \<Gamma>r \<subseteq>
                 \<Gamma>s,(z:Qual qL (TyRef \<tau>))" by auto
      have Hg7: "(\<Gamma>s,(z:Qual qL (TyRef \<tau>))) \<sim> (\<Gamma>m,(x:\<tau>))"
      proof (auto simp add: env_compat_def)
        show  "qualT \<tau> \<preceq> qU" using prems by auto
      next
        have "\<not> \<Gamma>m defines z" using Hsmnotin by (auto simp add: dom_def)
        thus "(z\<rhd>(Qual qL (TyRef \<tau>))) \<in> \<Gamma>m \<Longrightarrow> False"
          using Hnotin by (auto simp add: dom_def)
      next
        show "qualT \<tau> \<preceq> qU" using prems by auto
      next
        show "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>s; (x\<rhd>\<tau>) \<in> \<Gamma>m\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          using Hsmcomp by (auto simp add: env_compat_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{z\<rhd>Qual qL (TyRef \<tau>)}"],
               rule exI[of _ "\<Gamma>s,(z:Qual qL (TyRef \<tau>))"],
               rule exI[of _ "\<Gamma>m,(x:\<tau>)"], rule exI[of _ "M,(l:\<tau>)"])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_new_c by auto
    qed
  qed auto
next
  case (t_assign Mz \<Gamma>1 \<Gamma>2 \<tau> e1 e2 e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e'"  .
  thus ?case proof cases
    case (e_assign_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a)
    from t_assign
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_assign_p1 by auto
    moreover
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using t_assign env_compat_union_left by auto
    moreover
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK" using t_assign by auto
    moreover
    have "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_assign by auto
    moreover
    have "INV3 S" by assumption
    ultimately
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2"] by auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;{} \<turnstile> e1' := e2 : Qual qU TyUnit" proof
      show "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1' \<union> \<Gamma>2 \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" "\<Gamma>1' \<union> \<Gamma>2"] by auto
    next
      show "\<Gamma>1';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)" using Ht .
    next
      show "\<Gamma>2;{} \<turnstile> e2 : \<tau>" using t_assign by auto
    qed
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_assign env_compat_def by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_assign_p1 by auto
  next
    case (e_assign_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a) 
    from t_assign
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and> 
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';{} \<turnstile> e2' : \<tau>) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_assign_p2 by auto
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_assign env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    moreover
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK" using t_assign by auto
    moreover
    have "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_assign by auto
    moreover
    have "INV3 S" by assumption
    ultimately 
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';{} \<turnstile> e2' : \<tau>"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1"] by auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';{} \<turnstile> e1 := e2' : Qual qU TyUnit" proof
      show "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1 \<union> \<Gamma>2' \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" "\<Gamma>1 \<union> \<Gamma>2'"] by auto
    next
      show "\<Gamma>1;{} \<turnstile> e1 : Qual qU (TyRef \<tau>)" using t_assign by auto
    next 
      show "\<Gamma>2';{} \<turnstile> e2' : \<tau>" using Ht by auto 
    qed
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_assign env_compat_def by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto 
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M']) 
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_assign_p2 by auto
  next
    case (e_assign_c Sa Sa' \<mu>a l q x z)
    have Hsin: "x\<mapsto>TmLoc q l \<in> S" using e_assign_c by auto
    moreover
    have "x\<rhd>Qual qU (TyRef \<tau>) \<in> \<Gamma>1" using t_assign e_assign_c by auto
    hence Hxin: "x\<rhd>Qual qU (TyRef \<tau>) \<in> (\<Gamma>s\<union>\<Gamma>m)" using t_assign by auto
    moreover
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately
    have Hq: "q = qU" using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m"] by auto
    hence Hs: "S' = S" using e_assign_c by auto
    have "z\<rhd>\<tau> \<in> \<Gamma>2" using t_assign e_assign_c by auto
    hence Hin: "z\<rhd>\<tau> \<in> \<Gamma>s" using t_assign by auto
    have "M \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m)" using t_assign Hs by auto
    also have "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> \<Gamma>m,(z:\<tau>)" using Hin by auto
    finally
    have Hg1: "M \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m,(z:\<tau>))" using t_assign by auto
    moreover
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately
    obtain \<Gamma>v where "\<Gamma>v;M \<turnstile> TmLoc qU l : Qual qU (TyRef \<tau>)"
      using Hq Hxin Hsin storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc qU l" "Qual qU (TyRef \<tau>)"]
      by auto
    hence "l\<rhd>\<tau> \<in> M" by auto
    moreover
    hence "\<mu> defines l" using prems mem_dom_in[of \<Gamma>m \<mu> M l] by (auto simp add: dom_def)
    moreover have "\<Gamma>m,(z:\<tau>) \<Turnstile> OK" using e_assign_c t_assign wf_env_union_sanity_cons by auto
    moreover have "\<Gamma>m \<Turnstile>' \<mu> : M" by assumption
    moreover hence "\<Gamma>m \<sim> {z\<rhd>\<tau>}" by (auto simp add: env_compat_def)
    ultimately
    have Hg2: "\<Gamma>m,(z:\<tau>) \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
      using memTyping_update by auto
    have Hg3: "{};{} \<turnstile> TmUnit qU : Qual qU TyUnit" by (auto simp add: wf_env_def qualGrelL_def)
    have Hg4: "{} \<union> \<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
    have Hg5: "{} \<sim> \<Gamma>r" by (auto simp add: env_compat_def)
    hence Hg6: "{} \<union> \<Gamma>r \<Turnstile> OK"  using t_assign wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Gamma>r] by auto
    have "\<Gamma>s \<sim> \<Gamma>m" and "qualT \<tau> \<preceq> qU" by assumption
    hence Hg7: "\<Gamma>s \<sim> \<Gamma>m,(z:\<tau>)" by (auto simp add: env_compat_def)
    have Hg10: "INV2 S'" using Hs t_assign by auto
    have Hg11: "INV3 S'" using Hs t_assign by auto
    show ?thesis
      apply (rule exI[of _ "{}"], rule exI[of _ \<Gamma>s])
      apply (rule exI[of _ "\<Gamma>m,(z:\<tau>)"], rule exI[of _ M])
      using e_assign_c Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hs by auto
  qed auto
next
  case (t_swap Mz \<Gamma>1 \<Gamma>2 \<tau> e1 e2 e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e'"  .
  thus ?case proof cases
  case (e_swap_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a) 
    from t_swap
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and> 
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_swap_p1 by auto
    moreover
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using t_swap env_compat_union_left by auto
    moreover
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK" using t_swap by auto
    moreover
    have "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_swap by auto
    moreover
    have "INV3 S" by assumption
    ultimately
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2"] by auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;{} \<turnstile> e1' :=: e2 : \<tau>" proof
      show "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1' \<union> \<Gamma>2 \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" "\<Gamma>1' \<union> \<Gamma>2"] by auto
    next
      show "\<Gamma>1';{} \<turnstile> e1' : Qual qU (TyRef \<tau>)" using Ht .
    next 
      show "\<Gamma>2;{} \<turnstile> e2 : \<tau>" using t_swap by auto 
    qed
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_swap env_compat_def by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M']) 
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_swap_p1 by auto
  next
    case (e_swap_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a) 
    from t_swap
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and> 
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';{} \<turnstile> e2' : \<tau>) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (INV2 S') \<and> (INV3 S')" by auto
    have "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" and
      "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "Mz = {}" by assumption
    moreover
    have "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_swap_p2 by auto
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_swap env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    moreover
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK" using t_swap by auto
    moreover
    have "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_swap by auto
    moreover
    have "INV3 S" by assumption
    ultimately 
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';{} \<turnstile> e2' : \<tau>"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1"] by auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';{} \<turnstile> e1 :=: e2' : \<tau>" proof
      show "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto
    next
      show "\<Gamma>1 \<union> \<Gamma>2' \<Turnstile> OK"
        using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" "\<Gamma>1 \<union> \<Gamma>2'"] by auto
    next
      show "\<Gamma>1;{} \<turnstile> e1 : Qual qU (TyRef \<tau>)" using t_swap by auto
    next 
      show "\<Gamma>2';{} \<turnstile> e2' : \<tau>" using Ht by auto 
    qed
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_swap env_compat_def by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r \<Turnstile> OK" using H6_1 by auto 
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M']) 
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_swap_p2 by auto
  next
    case (e_swap_c Sa Sa' \<mu>a l q x z w)
    have Hsin: "x\<mapsto>TmLoc q l \<in> S" using e_swap_c by auto
    moreover
    have "x\<rhd>Qual qU (TyRef \<tau>) \<in> \<Gamma>1" using t_swap e_swap_c by auto
    hence Hxin: "x\<rhd>Qual qU (TyRef \<tau>) \<in> (\<Gamma>s\<union>\<Gamma>m)" using t_swap by auto
    moreover
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately 
    have Hq: "q = qU" using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m"] by auto 
    hence Hs: "S' = S" using e_swap_c by auto
    have Hzing2: "z\<rhd>\<tau> \<in> \<Gamma>2" using t_swap e_swap_c by auto
    hence Hin: "z\<rhd>\<tau> \<in> \<Gamma>s" using t_swap by auto
    have Hsttyp: "M \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m)" using t_swap Hs by auto
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    then obtain \<Gamma>v where "\<Gamma>v;M \<turnstile> TmLoc qU l : Qual qU (TyRef \<tau>)" 
      using Hq Hxin Hsin storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc qU l" "Qual qU (TyRef \<tau>)"]
      by auto
    hence HlinM: "l\<rhd>\<tau> \<in> M" by auto
    moreover
    hence Hlinm: "\<mu> defines l" using prems by (auto simp add: dom_def)
    ultimately
    have Hwingm: "w\<rhd>\<tau> \<in> \<Gamma>m" using t_swap e_swap_c memTyping_elim_qL[of \<Gamma>m \<mu> M l w \<tau>] by auto
    hence Hwnotings: "\<not> \<Gamma>s defines w" using t_swap compat_ok_sanity_right by auto 
    hence Hrnotw: "\<not> \<Gamma>r defines w" using t_swap by (auto simp add: dom_def)
    have "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>)" using t_swap Hin env_cons_without_eq_right by auto
    also
    {
      hence "\<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>) \<Turnstile> OK" using t_swap by auto
      moreover have "w\<rhd>\<tau> \<in> \<Gamma>m,(z:\<tau>)" using Hwingm by auto 
      ultimately
      have "\<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>) = (\<Gamma>s\<bar>z),(w:\<tau>) \<union> \<Gamma>m,(z:\<tau>)\<bar>w" 
        using env_cons_without_eq_left[of "\<Gamma>s\<bar>z" "\<Gamma>m,(z:\<tau>)" w \<tau>] by auto
    }
    also
    have Hneqzw: "z \<noteq> w" using Hin Hwnotings by (auto simp add: dom_def)
    hence Hgszw: "(\<Gamma>s\<bar>z),(w:\<tau>) = \<Gamma>s,(w:\<tau>)\<bar>z" by (auto simp add: without_def)
    finally
    have Heqenvs: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s,(w:\<tau>)\<bar>z \<union> \<Gamma>m,(z:\<tau>)\<bar>w" by auto
    hence Hg1: "M \<Turnstile> S':(\<Gamma>s,(w:\<tau>)\<bar>z \<union> \<Gamma>m,(z:\<tau>)\<bar>w)" using Hsttyp by auto
    have "\<Gamma>m \<Turnstile>' \<mu> : M" by assumption 
    moreover 
    have "\<not> \<Gamma>m defines z" using t_swap Hin compat_ok_sanity by auto
    ultimately
    have Hg2: "\<Gamma>m,(z:\<tau>)\<bar>w \<Turnstile>' update \<mu> l (TmFreeVar z) : M" 
      using t_swap e_swap_c Hneqzw HlinM Hwingm memTyping_update_qL_without[of \<Gamma>m \<mu> M l w \<tau> z]
      by auto
    have "z\<rhd>\<tau> \<in> (\<Gamma>1\<union>\<Gamma>2)" using Hzing2 by (auto simp add: dom_def)
    hence Hznotingr: "\<not> \<Gamma>r defines z" using t_swap compat_ok_sanity[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r z] by auto
    hence Hg3: "{(w\<rhd>\<tau>)};{} \<turnstile> TmFreeVar w : \<tau>"
      by (auto simp add: without_def wf_env_def qualGrelL_def) 
    have "{w\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> (\<Gamma>s\<bar>z),(w:\<tau>)" using t_swap Hznotingr 
      by (auto simp add: without_def dom_def)
    hence Hg4: "{w\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> \<Gamma>s,(w:\<tau>)\<bar>z" using Hgszw by auto
    have Hg5: "{w\<rhd>\<tau>} \<sim> \<Gamma>r" using Hrnotw by (auto simp add: dom_def env_compat_def)
    have "\<Gamma>r \<Turnstile> OK" using t_swap wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>r] by force
    hence Hg6: "{w\<rhd>\<tau>} \<union> \<Gamma>r \<Turnstile> OK" using t_swap Hrnotw wf_env_cons by auto
    have "\<Gamma>s \<sim> \<Gamma>m" by assumption 
    hence Hg7: "\<Gamma>s,(w:\<tau>)\<bar>z \<sim> \<Gamma>m,(z:\<tau>)\<bar>w" by (auto simp add: env_compat_def without_def)
    have Hg10: "INV2 S'" using Hs t_swap by auto
    have Hg11: "INV3 S'" using Hs t_swap by auto
    show ?thesis 
      apply (rule exI[of _ "{w\<rhd>\<tau>}"], rule exI[of _ "\<Gamma>s,(w:\<tau>)\<bar>z"])
      apply (rule exI[of _ "\<Gamma>m,(z:\<tau>)\<bar>w"], rule exI[of _ M])
      using e_swap_c Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hs by auto
  qed auto
next
  case (t_free Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_free_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_free
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
           (\<Gamma>e';{} \<turnstile> e' : Qual qL (TyRef \<tau>)) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (INV2 S') \<and> (INV3 S')" by auto
    have "Mz = {}" and "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'"
      and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
      and "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "INV3 S"
      using prems by auto
    moreover
    ultimately obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';{} \<turnstile> e' : Qual qL (TyRef \<tau>)"
      and H4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and H5: "\<Gamma>e' \<sim> \<Gamma>r"
      and H6: "\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r] by auto
    moreover
    have Hg3: "\<Gamma>e';{} \<turnstile> TmFree e' : \<tau>" using Ht by auto
    ultimately show ?thesis using e_free_p by auto
  next
    case (e_free_c Sa Sa' \<mu>a l q x y)
    have Hx: "x\<rhd>Qual qL (TyRef \<tau>) \<in> \<Gamma>" using t_free e_free_c by auto
    hence Hsx: "x\<rhd>Qual qL (TyRef \<tau>) \<in> \<Gamma>s" using t_free by auto
    hence "x\<rhd>Qual qL (TyRef \<tau>) \<in> (\<Gamma>s \<union> \<Gamma>m)" by auto
    moreover
    have HxinS: "(x\<mapsto>TmLoc q l) \<in> S" using t_free e_free_c by auto
    moreover
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately
    obtain \<Gamma>v
      where Ht: "\<Gamma>v;M \<turnstile> TmLoc q l : Qual qL (TyRef \<tau>)"
      and Hst: "M \<Turnstile> S\<bar>x : ((\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v)"
      and Hcomp: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<sim> \<Gamma>v"
      and Hok: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v \<Turnstile> OK"
      using storeTyping_elim[of M S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q l" "Qual qL (TyRef \<tau>)"]
      by auto
    have Hsok: "\<Gamma>s \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>s] by auto
    have Hascomp: "\<Gamma>s \<sim> \<Gamma>m" and Hasok: "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" by assumption
    hence Hmx: "\<not> \<Gamma>m defines x" using Hsx compat_ok_sanity by auto
    have "l\<rhd>\<tau> \<in> M" and "l\<mapsto>TmFreeVar y \<in> \<mu>" and "\<Gamma>m \<Turnstile>' \<mu> : M"
      using prems Ht by auto
    hence Hmt1: "\<lbrakk> qualT \<tau> = qL \<rbrakk> \<Longrightarrow> \<Gamma>m\<bar>y \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (y\<rhd>\<tau>) \<in> \<Gamma>m"
      and Hmt2: "\<lbrakk> qualT \<tau> = qU \<rbrakk> \<Longrightarrow> \<Gamma>m \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (y\<rhd>\<tau>) \<in> \<Gamma>m"
      using memTyping_elim_qL memTyping_elim_qU by auto
    hence Hmy: "y\<rhd>\<tau> \<in> \<Gamma>m" by (cases "qualT \<tau>", auto)
    have Hql: "q = qL" using Ht by auto
    hence Hs': "S\<bar>x = S'" using e_free_c by auto
    have "\<not> l free loc in st (S\<bar>x)" using t_free HxinS Hql by auto
    hence Hg1u: "M\<bar>l \<Turnstile> S\<bar>x : (((\<Gamma>s\<union>\<Gamma>m)\<bar>x) \<union> \<Gamma>v)" using Hst storeTyping_strengthen by auto
    also have " (\<Gamma>s\<union>\<Gamma>m)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x) \<union> \<Gamma>v \<union> \<Gamma>m"
      using Hmx by (auto simp add: without_def dom_def)
    also have "(\<Gamma>s\<bar>x) \<union> \<Gamma>v \<union> \<Gamma>m = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m" using Hmy by auto
    finally have "M\<bar>l \<Turnstile> S\<bar>x : ((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)" by auto
    hence Hg1u: "M\<bar>l \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)" using Hs' by auto
    have Hg3: "{(y\<rhd>\<tau>)};{} \<turnstile> TmFreeVar y : \<tau>"
      by (auto simp add: wf_env_def qualGrelL_def without_def)
    have Hasrs: "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
    moreover
    have "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r \<Turnstile> OK" by assumption
    hence "\<not> \<Gamma>r defines x" using Hx compat_ok_sanity by auto
    ultimately
    have "\<Gamma>r \<subseteq> (\<Gamma>s\<bar>x)" by (auto simp add: without_def dom_def)
    hence "\<Gamma>r \<subseteq> (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v" by auto
    hence Hg4: "{(y\<rhd>\<tau>)} \<union> \<Gamma>r \<subseteq> (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v" by auto
    have Hrnotin: "\<not> \<Gamma>r defines y \<or> (y\<rhd>\<tau> \<in> \<Gamma>r \<and> qualT \<tau> = qU)"
    proof -
      {
        assume Hry: "\<Gamma>r defines y"
        hence "\<Gamma>s defines y" using prems by (auto simp add: dom_def)
        hence Hsy: "y\<rhd>\<tau> \<in> \<Gamma>s"
          using prems Hmy wf_env_union_sanity[of \<Gamma>s \<Gamma>m] by auto
        hence "y\<rhd>\<tau> \<in> \<Gamma>r" using Hasrs Hsok Hry wf_env_subset_defines by auto
        moreover
        {
          assume Hl: "qualT \<tau> = qL"
          hence "\<not> \<Gamma>m defines y" using Hascomp Hasok Hsy compat_ok_sanity
            by auto
          hence False using Hmy by (auto simp add: dom_def)
        }
        hence "qualT \<tau> = qU" by (cases "qualT \<tau>", auto)
        ultimately
        have "y\<rhd>\<tau> \<in> \<Gamma>r \<and> qualT \<tau> = qU" by auto
      }
      thus ?thesis by auto
    qed
    have Hg5: "{(y\<rhd>\<tau>)} \<sim> \<Gamma>r"
      using Hrnotin by (auto simp add: dom_def env_compat_def)
    have "\<Gamma>r \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>\<union>\<Gamma>r" \<Gamma>r] by auto
    hence Hg6: "{(y\<rhd>\<tau>)} \<union> \<Gamma>r \<Turnstile> OK"
      using Hrnotin by (auto simp add: dom_def wf_env_def)
    have Hs: "S\<bar>x \<subseteq> S" by (auto simp add: without_def)
    moreover
    have "INV2 S" by assumption
    ultimately
    have Hg10: "INV2 (S\<bar>x)" using inv2_subset by auto
    have "INV3 S" by assumption
    hence Hg11: "INV3 (S\<bar>x)" using Hs inv3_subset by auto
    show ?thesis proof (cases "qualT \<tau>")
      case qL
      note Hst
      also
      have Heql: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y" proof -
        have Heq: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x = \<Gamma>s\<bar>x \<union> \<Gamma>m" using Hmx
          by (auto simp add: dom_def without_def)
        also
        {
          have "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<Turnstile> OK" using Hok wf_env_subset by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m \<Turnstile> OK" using Heq by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>m\<bar>y"
            using Hmy env_cons_without_eq_left[of "\<Gamma>s\<bar>x" \<Gamma>m] by auto
        }
        finally
        show ?thesis by auto
      qed
      also note Hs'
      finally
      have "M \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y)" by auto
      moreover
      hence "\<not> l free loc in st (S\<bar>x)" using HxinS Hql t_free by auto 
      hence "\<not> l free loc in st S'" using Hs' by auto
      ultimately
      have Hg1l: "M\<bar>l \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y)" using storeTyping_strengthen by auto
      have Hg2: "\<Gamma>m\<bar>y \<Turnstile>' \<mu>':M\<bar>l" using e_free_c Hmt1 qL by auto
      have Hg7: "((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v) \<sim> \<Gamma>m\<bar>y" using Hcomp
      proof (auto simp add: env_compat_def)
        assume "(y\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y"
        hence False by (auto simp add: dom_def without_def)
        thus "qualT \<tau> \<preceq> qU" by auto
      next
        fix xa \<tau>
        assume "(xa\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y"
          and "(xa\<rhd>\<tau>) \<in> \<Gamma>s\<bar>x"
        moreover
        have "\<Gamma>s \<sim> \<Gamma>m" by assumption
        ultimately
        show "qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      next
        have "\<Gamma>m\<bar>x \<sim> \<Gamma>v"
          using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>x" \<Gamma>v "\<Gamma>m\<bar>x"]
          by (auto simp add: without_def)
        also have "\<Gamma>m\<bar>x = \<Gamma>m" using Hmx by (auto simp add: without_def dom_def)
        finally have "\<Gamma>m \<sim> \<Gamma>v" by auto
        thus "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y; (x\<rhd>\<tau>) \<in> \<Gamma>v\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{y\<rhd>\<tau>}"],
          rule exI[of _ "(\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v"],
          rule exI[of _ "\<Gamma>m\<bar>y"], rule exI[of _ "M\<bar>l"])
        using Hg1l Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_free_c Hs' by auto
    next
      case qU
      note Hst
      also
      have HeqU: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m" proof -
        have Heq: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x = \<Gamma>s\<bar>x \<union> \<Gamma>m" using Hmx
          by (auto simp add: dom_def without_def)
        also
        {
          have "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<Turnstile> OK" using Hok wf_env_subset by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m \<Turnstile> OK" using Heq by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>m"
            using Hmy env_cons_without_eq_left[of "\<Gamma>s\<bar>x" \<Gamma>m] by auto
        }
        finally
        show ?thesis by auto
      qed
      also note Hs'
      finally
      have "M \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)" by auto
      moreover
      hence "\<not> l free loc in st (S\<bar>x)" using HxinS Hql t_free by auto 
      hence "\<not> l free loc in st S'" using Hs' by auto
      ultimately
      have Hg1U: "M\<bar>l \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)" using storeTyping_strengthen by auto
      have Hg2: "\<Gamma>m \<Turnstile>' \<mu>':M\<bar>l" using e_free_c Hmt2 qU by auto
      have Hg7: "((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v) \<sim> \<Gamma>m"
      proof (auto simp add: env_compat_def)
        assume "(y\<rhd>\<tau>) \<in> \<Gamma>m"
        thus "qualT \<tau> \<preceq> qU" using qU by auto
      next
        fix xa \<tau>
        assume "(xa\<rhd>\<tau>) \<in> \<Gamma>m"
          and "(xa\<rhd>\<tau>) \<in> \<Gamma>s\<bar>x"
        moreover
        have "\<Gamma>s \<sim> \<Gamma>m" by assumption
        ultimately
        show "qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      next
        have "\<Gamma>m\<bar>x \<sim> \<Gamma>v"
          using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>x" \<Gamma>v "\<Gamma>m\<bar>x"]
          by (auto simp add: without_def)
        also have "\<Gamma>m\<bar>x = \<Gamma>m" using Hmx by (auto simp add: without_def dom_def)
        finally have "\<Gamma>m \<sim> \<Gamma>v" by auto
        thus "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>m; (x\<rhd>\<tau>) \<in> \<Gamma>v\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{y\<rhd>\<tau>}"],
               rule exI[of _ "(\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v"],
               rule exI[of _ "\<Gamma>m"], rule exI[of _ "M\<bar>l"])
        using Hg1u Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_free_c Hs' by auto
    qed
  qed auto
next
  case (t_deref Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_deref_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_deref
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m \<Turnstile> OK;
           INV2 S; INV3 S\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M'.
            (M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m' \<Turnstile>' \<mu>':M') \<and>
           (\<Gamma>e';{} \<turnstile> e' : Qual qU (TyRef \<tau>)) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (INV2 S') \<and> (INV3 S')" by auto
    have "Mz = {}" and "M \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m \<Turnstile>' \<mu>:M" and "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'"
      and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
      and "\<Gamma>s \<union> \<Gamma>m \<Turnstile> OK" and "INV2 S" and "INV3 S"
      using prems by auto
    moreover
    ultimately obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M'
      where Hg1: "M' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m' \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';{} \<turnstile> e' : Qual qU (TyRef \<tau>)"
      and H4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and H5: "\<Gamma>e' \<sim> \<Gamma>r"
      and H6: "\<Gamma>e' \<union> \<Gamma>r \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 S'"
      and Hg11: "INV3 S'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r] by auto
    moreover
    have Hg3: "\<Gamma>e';{} \<turnstile> TmDeref e' : \<tau>" using Ht by auto
    ultimately show ?thesis using e_deref_p by auto
  next
    case (e_deref_c Sa Sa' \<mu>a l q x y)
    have Hx: "x\<rhd>Qual qU (TyRef \<tau>) \<in> \<Gamma>" using t_deref e_deref_c by auto
    hence Hsx: "x\<rhd>Qual qU (TyRef \<tau>) \<in> \<Gamma>s" using t_deref by auto
    hence "x\<rhd>Qual qU (TyRef \<tau>) \<in> (\<Gamma>s \<union> \<Gamma>m)" by auto
    moreover
    have HxinS: "(x\<mapsto>TmLoc q l) \<in> S" using t_deref e_deref_c by auto
    moreover
    have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately
    obtain \<Gamma>v
      where Ht: "\<Gamma>v;M \<turnstile> TmLoc q l : Qual qU (TyRef \<tau>) "
      and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>m)"
      and HqU: "q = qU"
      using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q l" "Qual qU (TyRef \<tau>)"]
      by auto
    hence Hlinm: "l\<rhd>\<tau> \<in> M" by auto
    have Hsok: "\<Gamma>s \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Gamma>s] by auto
    have Hs: "S' = S" using HqU e_deref_c by auto
    have "qualT \<tau> = qU" using qualT_unrestricted t_deref by auto
    hence Hyingm: "y\<rhd>\<tau> \<in> \<Gamma>m" 
      using Hlinm t_deref e_deref_c memTyping_elim_qU[of \<Gamma>m \<mu> M l y \<tau>] by auto
    hence Hgeq: "\<Gamma>s,(y:\<tau>) \<union> \<Gamma>m = \<Gamma>s \<union> \<Gamma>m" by auto
    hence Hg1: "M \<Turnstile> S:(\<Gamma>s,(y:\<tau>) \<union> \<Gamma>m)" using t_deref by auto
    have Hg2: "\<Gamma>m \<Turnstile>' \<mu> : M" by assumption
    have Hg3: "{y\<rhd>\<tau>};{} \<turnstile> TmFreeVar y : \<tau>"
      by (auto simp add: wf_env_def without_def qualGrelL_def)
    have Hg4: "{y\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> \<Gamma>s,(y:\<tau>)" using t_deref by auto
    have Hg5: "{y\<rhd>\<tau>} \<sim> \<Gamma>r" using t_deref by (auto simp add: env_compat_def)
    have "\<Gamma>r \<Turnstile> OK" using Hsok t_deref wf_env_subset[of "\<Gamma>s" \<Gamma>r] by auto
    have "\<Gamma>r,(y:\<tau>) \<subseteq> \<Gamma>s\<union>\<Gamma>m" using t_deref Hyingm by auto
    hence Hg6: "{y\<rhd>\<tau>} \<union> \<Gamma>r \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" "{y\<rhd>\<tau>} \<union> \<Gamma>r"] t_deref by auto
    have Hg7: "\<Gamma>s,(y:\<tau>) \<sim> \<Gamma>m" using t_deref by (auto simp add: env_compat_def)
    have Hg10: "INV2 S" by assumption
    have Hg11: "INV3 S" by assumption
    show ?thesis
      apply (rule exI[of _ "{y\<rhd>\<tau>}"],
        rule exI[of _ "\<Gamma>s,(y:\<tau>)"],
        rule exI[of _ "\<Gamma>m"], rule exI[of _ "M"])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 e_deref_c Hs by auto
    qed auto
  qed
qed

theorem progress:
  assumes "\<Gamma>e; {} \<turnstile> e : \<tau>"
  and "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)"
  and "\<Gamma>m \<Turnstile>' \<mu> : M"
  and "\<Gamma>e \<subseteq> \<Gamma>s"
  and "INV2 S"
  and "INV3 S"
  shows "not_stuck e S \<mu>"
proof -
  { fix Me
    assume "\<Gamma>e; Me \<turnstile> e : \<tau>"
      and "Me = {}"
      and "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)"
      and "\<Gamma>m \<Turnstile>' \<mu> : M"
      and "\<Gamma>e \<subseteq> \<Gamma>s"
      and "INV2 S"
      and "INV3 S"
    moreover have "S \<Turnstile> Store" using prems storeTyping_wf_store by auto 
    moreover have "\<mu> \<Turnstile> Mem" using prems memTyping_wf_mem by auto
    moreover have "\<forall>l. \<not> l free loc in e" using prems ty_no_locs by auto
    ultimately
    have "not_stuck e S \<mu>" proof induct
      case (t_unit M \<Gamma> q)
      have "value (TmUnit q)" by auto
      moreover
      have "finite (dom S)" 
        using prems storeTyping_wf_store[of _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S"
        using prems fresh_exists[of "dom S"] by auto
      ultimately
      have "S;\<mu>;TmUnit q \<hookrightarrow> S,(z\<mapsto>TmUnit q);\<mu>;TmFreeVar z" using t_unit by auto
      thus ?case by force
    next
      case t_var thus ?case by auto
    next
      case (t_abs L M \<Gamma> \<tau>1 \<tau>2 e_b q)
      have "value (\<lambda>^q[\<tau>1]. e_b)" by auto
      moreover
      have "finite (dom S)" 
        using prems storeTyping_wf_store[of _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S"
        using prems fresh_exists[of "dom S"] by auto
      ultimately
      have "S;\<mu>;(\<lambda>^q[\<tau>1]. e_b) \<hookrightarrow> S,(z\<mapsto>(\<lambda>^q[\<tau>1]. e_b));\<mu>;TmFreeVar z"
        using t_abs by auto
      thus ?case by auto
    next
      case (t_app Ma \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q)
      moreover
      have "\<forall>l. \<not> l free loc in e1" using t_app by auto
      ultimately
      have "isVar e1 \<or> makes_progress e1 S \<mu>" using t_app by auto
      thus ?case proof
        assume V1: "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by auto
        have "\<forall>l. \<not> l free loc in e2" using t_app by auto
        hence "isVar e2 \<or> makes_progress e2 S \<mu>" using t_app by auto
        thus ?case proof
          assume V2: "isVar e2" 
          then obtain y where Hy: "e2 = TmFreeVar y" by auto
          have "x \<in> dom \<Gamma>s" using Hx t_app by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_app by (auto simp add: store_dom_in)
          moreover have Hstore: "S \<Turnstile> Store" by assumption
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
            using store_ok_exists_value by auto
          moreover
          have "(x\<rhd>Qual q (\<tau>1 \<rightarrow> \<tau>2)) \<in> \<Gamma>1" using Hx t_app by auto
          hence "(x\<rhd>Qual q (\<tau>1 \<rightarrow> \<tau>2)) \<in> \<Gamma>s" using t_app by auto
          moreover
          have "M \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
          ultimately
          obtain \<Gamma>' where "\<Gamma>';M \<turnstile> v : Qual q (\<tau>1 \<rightarrow> \<tau>2)"
            using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x v] Hstore by blast
          then obtain tm where "v = \<lambda>^q[\<tau>1]. tm" using Hv by (cases v, auto)
          then obtain S' where "S;x\<Down>S';\<lambda>^q[\<tau>1].tm" using Hsin by auto
          hence H: "S;\<mu>;TmFreeVar x \<cdot> TmFreeVar y \<hookrightarrow> S';\<mu>;freshen_tm y tm"
            using t_app by blast
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x \<cdot> TmFreeVar y \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx Hy by auto
        next
          assume "makes_progress e2 S \<mu>"
          then obtain e2' S' \<mu>' where "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(e1 \<cdot> e2) \<hookrightarrow> S';\<mu>';(e1 \<cdot> e2')" using V1 by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>; e1\<cdot>e2 \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx by auto
        qed
      next
        assume "makes_progress e1 S \<mu>"
        then obtain e1' S' \<mu>' where "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;(e1 \<cdot> e2) \<hookrightarrow> S';\<mu>';(e1' \<cdot> e2)" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; e1\<cdot>e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    next
      case (t_new Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_new by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_new by auto
      thus ?case proof
        assume "isVar e"
        obtain x where Hx: "e = TmFreeVar x" by (rule isVar_var_elim)
        have "finite (dom S)" 
          using prems storeTyping_wf_store[of _ S] wf_store_def[of S] dom_finite[of S] by auto
        then obtain z where Hz: "\<not> S defines z" using fresh_exists[of "dom S"]
          by (auto simp add: dom_finite)
        have "finite (dom \<mu>)" 
          using memTyping_wf_mem[of _ \<mu>] wf_mem_def[of \<mu>] dom_finite[of \<mu>] prems by auto
        then obtain l where Hl: "\<not> \<mu> defines l" using fresh_exists[of "dom \<mu>"]
          by (auto simp add: dom_finite)
        have "\<Gamma>s defines x" using prems by (auto simp add: dom_def)
        hence "(\<Gamma>s\<union>\<Gamma>m) defines x" by (auto simp add: dom_def)
        hence "S defines x" using prems by (auto simp add: store_dom_in)
        moreover have "S \<Turnstile> Store" by assumption
        ultimately
        obtain v where "(x\<mapsto>v) \<in> S"
          using prems store_ok_exists_value[of S x] by auto
        hence H: "S;\<mu>;TmNew (TmFreeVar x) \<hookrightarrow>
               S,(z\<mapsto>TmLoc qL l);\<mu>,(l\<mapsto>TmFreeVar x);TmFreeVar z"
          using t_new Hx Hz Hl e_new_c[of S \<mu> l z] by blast
        have "\<exists>e' S' \<mu>'. S;\<mu>; TmNew (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';TmNew e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; TmNew e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case by auto
      qed
    next
      case (t_assign Ma \<Gamma>1 \<Gamma>2 \<tau> e1 e2)
      moreover have "\<forall>l. \<not> l free loc in e1" using t_assign by auto
      ultimately have "isVar e1 \<or> makes_progress e1 S \<mu>" using t_assign by auto
      thus ?case proof
        assume "makes_progress e1 S \<mu>"
        then obtain e1' S' \<mu>' where "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e1' := e2" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; e1:=e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto 
      next
        assume "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by (rule isVar_var_elim)
        moreover have "\<forall>l. \<not> l free loc in e2" using t_assign by auto
        ultimately have "isVar e2 \<or> makes_progress e2 S \<mu>" using t_assign by auto
        thus ?case proof
          assume "makes_progress e2 S \<mu>"
          then obtain e2' S' \<mu>' where "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(TmFreeVar x) := e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) := e2'" by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x := e2 \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?case using Hx by auto
        next
          assume "isVar e2"
          obtain y where Hy: "e2 = TmFreeVar y" by (rule isVar_var_elim)
          have Hstore: "S \<Turnstile> Store" by assumption
          moreover
          have "x \<in> dom \<Gamma>s" using Hx t_assign by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_assign store_dom_in by auto
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
            using store_ok_exists_value by auto
          then obtain S' where Hd1: "S;x\<Down>S';v" by auto
          moreover
          have "(x\<rhd>Qual qU (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_assign by auto
          ultimately
          obtain \<Gamma>' where "\<Gamma>';M \<turnstile> v : Qual qU (TyRef \<tau>)"
            using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x v] t_assign Hsin by blast
          then obtain l where Hv: "v = TmLoc qU l"
            using Hv by (cases v, auto)
          have H:"S;\<mu>;TmAssign (TmFreeVar x) (TmFreeVar y) \<hookrightarrow>
                            S';update \<mu> l (TmFreeVar y); TmUnit qU"
            using t_assign Hd1 Hv by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x := TmFreeVar y \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx Hy by auto  
        qed
      qed
    next
      case (t_swap Ma \<Gamma>1 \<Gamma>2 \<tau> e1 e2)
      moreover have "\<forall>l. \<not> l free loc in e1" using t_swap by auto
      ultimately have "isVar e1 \<or> makes_progress e1 S \<mu>" using t_swap by auto
      thus ?case proof
        assume "makes_progress e1 S \<mu>"
        then obtain e1' S' \<mu>' where "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e1' :=: e2" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; e1 :=: e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case by auto
      next
        assume "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by (rule isVar_var_elim)
        moreover have "\<forall>l. \<not> l free loc in e2" using t_swap by auto
        ultimately have "isVar e2 \<or> makes_progress e2 S \<mu>" using t_swap by auto
        thus ?case proof
          assume "makes_progress e2 S \<mu>"
          then obtain e2' S' \<mu>' where "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(TmFreeVar x) :=: e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) :=: e2'" by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x :=: e2 \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?case using Hx by auto
        next
          assume "isVar e2"
          obtain y where Hy: "e2 = TmFreeVar y" by (rule isVar_var_elim)
          have Hstore: "S \<Turnstile> Store" by assumption
          moreover
          have "x \<in> dom \<Gamma>s" using Hx t_swap by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_swap store_dom_in by auto
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v" 
            using store_ok_exists_value by auto
          then obtain S' where Hd1: "S;x\<Down>S';v" by auto
          moreover
          have "(x\<rhd>Qual qU (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_swap by auto
          ultimately
          obtain \<Gamma>' where "\<Gamma>';M \<turnstile> v : Qual qU (TyRef \<tau>)"
            using storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x v] t_swap Hsin by blast
          then obtain l where Hv: "v = TmLoc qU l"  and "l\<rhd>\<tau> \<in> M" 
            using Hv by (cases v, auto)
          moreover          
          hence "\<mu> defines l" using t_swap mem_dom_in[of \<Gamma>m \<mu> M l] by (auto simp add: dom_def)
          then obtain z where "(l\<mapsto>TmFreeVar z) \<in> \<mu>" 
            using t_swap mem_ok_exists_var by blast
          ultimately
          have H: "S;\<mu>;(TmFreeVar x) :=: (TmFreeVar y) \<hookrightarrow>
                            S';update \<mu> l (TmFreeVar y); TmFreeVar z"
            using t_swap Hd1 Hv by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x :=: TmFreeVar y \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx Hy by auto  
        qed
      qed
    next
      case (t_deref Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_deref by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_deref by auto
      thus ?case proof
        assume "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have "x \<in> dom \<Gamma>s" using Hx t_deref by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_deref store_dom_in by auto
        then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hval: "value v"
          using store_ok_exists_value t_deref by blast
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have "(x\<rhd>Qual qU (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_deref by auto
        then obtain \<Gamma>' where Htm: "\<Gamma>';M \<turnstile> v : Qual qU (TyRef \<tau>)"
          using t_deref Hsin storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x v] by best
        then obtain l where Hv: "v = TmLoc qU l"
          using Hval by (cases v, auto)
        hence "l \<in> dom M" using Htm by (auto simp add: dom_def)
        hence "l \<in> dom \<mu>" using mem_dom_in[of \<Gamma>m \<mu> M l] t_deref by auto
        then obtain z where "(l\<mapsto>TmFreeVar z) \<in> \<mu>" 
          using t_deref mem_ok_exists_var by blast
        hence H: "S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow> S';\<mu>;TmFreeVar z"
          using t_deref Hs' Hv by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';TmDeref e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    next
      case (t_free Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_free by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_free by auto
      thus ?case proof
        assume "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have "x \<in> dom \<Gamma>s" using Hx t_free by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_free store_dom_in by auto
        then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hval: "value v"
          using store_ok_exists_value t_free by blast
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have "(x\<rhd>Qual qL (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_free by auto
        then obtain \<Gamma>' where Htm: "\<Gamma>';M \<turnstile> v : Qual qL (TyRef \<tau>)"
          using t_free Hsin storeTyping_sanity[of M S "\<Gamma>s\<union>\<Gamma>m" x v] by best
        then obtain l where Hv: "v = TmLoc qL l"
          using Hval by (cases v, auto)
        hence "l \<in> dom M" using Htm by (auto simp add: dom_def)
        hence "l \<in> dom \<mu>" using mem_dom_in[of \<Gamma>m \<mu> M l] t_free by auto
        then obtain y where "(l\<mapsto>TmFreeVar y) \<in> \<mu>"
          using t_free mem_ok_exists_var by blast
        hence H: "S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>\<bar>l;TmFreeVar y"
          using t_free Hs' Hv e_free_c[of S \<mu>] by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';TmFree e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    next
      case (t_loc Ma \<Gamma> \<tau> l q)
      have "value (TmLoc q l)" by auto
      moreover
      have "finite (dom S)" 
        using prems storeTyping_wf_store[of _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S" using fresh_exists[of "dom S"] by auto
      ultimately
      have H: "S;\<mu>;(TmLoc q l) \<hookrightarrow> S,(z\<mapsto>(TmLoc q l));\<mu>;TmFreeVar z"
        using t_loc by auto
      have "\<exists>e' S' \<mu>'. S;\<mu>;TmLoc q l \<hookrightarrow> S';\<mu>';e'"
        apply (rule exI)+ using H by auto
      thus ?case by auto
    qed
  }
  with prems show ?thesis by auto
qed

theorem safety:
  assumes Ht: "\<Gamma>e;{} \<turnstile> e : \<tau>"
  and Hsem: "S;\<mu>;e \<hookrightarrow>* S';\<mu>';e'"
  and H1: "M \<Turnstile> S : \<Gamma>s \<union> \<Gamma>m"
  and H2: "\<Gamma>m \<Turnstile>' \<mu> : M"
  and H3: "\<Gamma>e \<subseteq> \<Gamma>s"
  and H4: "\<Gamma>s \<sim> \<Gamma>m"
  and H5: "INV2 S"
  and H6: "INV3 S"
  shows "not_stuck e' S' \<mu>'"
using Hsem Ht H1 H2 H3 H4 H5 H6 proof (induct arbitrary: \<Gamma>e \<Gamma>s \<Gamma>m M)
  case e_refl thus ?case using progress by auto
next
  case (e_trans S S' S'' \<mu> \<mu>' \<mu>'' e e' e'' \<Gamma>e \<Gamma>s \<Gamma>m M)
  have "\<Gamma>e \<Turnstile> OK" using prems by (auto simp add: ty_wf_env)
  hence "\<Gamma>e \<union> {} \<Turnstile> OK" and "\<Gamma>e \<sim> {}" by (auto simp add: env_compat_def)
  then obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' where "\<Gamma>e';{} \<turnstile> e' : \<tau>"
    and "M' \<Turnstile> S' : (\<Gamma>s'\<union>\<Gamma>m')"
    and "\<Gamma>e' \<subseteq> \<Gamma>s'"
    and "\<Gamma>m' \<Turnstile>' \<mu>' : M'"
    and "\<Gamma>s' \<sim> \<Gamma>m'"
    and "INV2 S'"
    and "INV3 S'"
    using preservation[of \<Gamma>e "{}" e \<tau> M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' e' "{}"] using e_trans by auto 
  thus ?case using progress e_trans by auto
qed

theorem program_safety:
  assumes "{};{} \<turnstile> e : \<tau>"
  and "{};{};e \<hookrightarrow>* S';\<mu>';e'"
  shows "not_stuck e' S' \<mu>'"
using prems by (rule safety, auto simp add: wf_env_def env_compat_def)

end
